'''
Created on 18.09.2019

@author: Samuel
'''
import sys
from itertools import repeat, starmap
from multiprocessing import Pool
from collections import defaultdict
from functools import partial

import numpy as np
from scipy import optimize as op, integrate, sparse
from scipy.stats import chi2, nbinom

try:
    from matplotlib import pyplot as plt
except ModuleNotFoundError:
    raise ModuleNotFoundError("The package matplotlib must be installed to run the test.")

try:
    from numdifftools import Gradient, Hessian
    import autograd.numpy as ag
except ModuleNotFoundError:
    raise ModuleNotFoundError("This module requires numdifftools and autograd "
                              "for numerical and automatic differenciation. "
                              "The packages are available on the Python package"
                              "index.")

try:
    from .ci_rvm import find_CI_bound, CounterFun, find_CI as find_CI_new, find_function_CI
    from .__ci_rvm_mpinv import find_profile_CI_bound as find_profile_CI_bound_mpinv
except ImportError:
    from ci_rvm import find_CI_bound, CounterFun, find_CI as find_CI_new, find_function_CI
    from __ci_rvm_mpinv import find_profile_CI_bound as find_profile_CI_bound_mpinv

if __name__ == '__main__':
    try:
        # if vemomoto_core_tools is installed, you can call the module with an 
        # argument specifying a log file to which the output is written in addition
        # to the console.
        from vemomoto_core.tools.tee import Tee
        if len(sys.argv) > 1:
            teeObject = Tee(sys.argv[1])
    except ModuleNotFoundError:
        pass

np.random.seed()

CONVERT = True
ALPHA = 0.95
def convertPos(x, indices=None, convert=CONVERT):
    if not convert:
        return x
    x = x.copy()
    
    """
    res = x[indices]
    cons = res > 50
    if not cons.any():
        res[:] = np.log(np.exp(res)+1)
    else: 
        res2 = np.log(np.exp(res)+1)
        res2[cons] = res[cons]
        res = res2
    x[indices] = res
    return x
    """
    
    x[indices] = np.exp(x[indices])
    return x

def convertPos_(x, convert=CONVERT):
    if not convert:
        return x
    return np.exp(x) #!!!!!!!
    if x <= 50:
        return np.log(np.exp(x)+1)
    else: 
        return x

def convertPosInv(x, indices=None, convert=CONVERT):
    if not convert:
        return x
    x = x.copy()
    res = x[indices]
    """
    cons = res > 50
    res2 = np.log(np.exp(res)-1)
    if cons.any():
        res2[cons] = res[cons]
    """
    res2 = np.log(res) #!!!!!!!!
    res2[~np.isfinite(res2)] = -10 #000
    x[indices] = res2
    return x

def convertPosInv_(x, convert=CONVERT):
    if not convert:
        return x
    res = np.log(x)
    if not np.isfinite(res):
        res = -10
    return res
    
    
    if not convert or x > 50:
        return x
    res = np.log(np.exp(x)-1)
    if not np.isfinite(res):
        res = -10000
    return res

class TrackFun():
    def __init__(self, fun):
        self.track = []
        self.fun = fun
    def __call__(self, args):
        self.track.append(args.copy())
        return self.fun(args)


def venzon_moolgavkar(index, direction, x0, fun, jac, hess, target=None, alpha=0.95, 
                      fun0=None, hess0=None, nmax=200, epsilon=1e-6, disp=True,
                      track_x=False):
    
    if fun0 is None:
        fun0 = fun(x0)
    
    if target is None:
        diff = chi2.ppf(alpha, df=1)/2
        target = fun0-diff
    
    step_scale = direction
    
    x_track = []
    
    # TODO: handle singular matrices (check determinant, return NaN), extend algorithm to be more robust
    i = 0
    try:    
        # preparation
        
        if fun0 is None:
            fun0 = fun(x0)
        target_diff = fun0-target
        if hess0 is None:
            dl2_dx2_0 = hess(x0)
        else:
            dl2_dx2_0 = hess0
        
        considered = np.ones_like(x0, dtype=bool)
        considered[index] = False
        hessConsidered = np.ix_(considered, considered)
        
        # test
        if False:
            f = lambda x: (np.square(fun(x)-target) 
                           + np.sum(np.square(jac(x)[considered])))
            print(op.minimize(f, x0))
                           
        # x is the parameter vector
        # w is the vector of nuisance parameters (x without row "index")
        # b is the parameter of interest, which is x[index]
        # l is fun
        # dl_d* is the derivative of l w.r.t. *
        # dl2_d*2 is the second derivative of l w.r.t. *
        
        # choosing the first step x1
        
        dl2_dw2 = dl2_dx2_0[hessConsidered]
        
        dl2_dbdw = dl2_dx2_0[index][considered]
        dl2_dw2_inv = np.linalg.inv(dl2_dw2)
        dw_db = -np.dot(dl2_dw2_inv, dl2_dbdw)
        
        factor = (np.sqrt(target_diff / (-2*(dl2_dx2_0[index, index]
                            - np.dot(np.dot(dl2_dbdw, dl2_dw2_inv), dl2_dbdw))))
                                               * step_scale)
        
        if np.isnan(factor):
            factor = 0.5
            #raise np.linalg.LinAlgError()
        
        init_direction = np.zeros_like(x0)
        init_direction[considered] = dw_db
        init_direction[index] = 1
        x = x0 + factor*init_direction
        
        
        # iteration
        for i in range(nmax):
            if track_x:
                x_track.append(x.copy())
            l = fun(x)
            dl_dx_ = jac(x)
            violation = max(np.max(np.abs(dl_dx_[considered])), np.abs(l-target))
            
            if disp:
                #print(x)
                print("iteration {}: fun_diff={}; jac_violation={}; direction={}".format(
                        i, l-target, np.max(np.abs(dl_dx_[considered])), step_scale))
            
            if violation < epsilon:
                break
            elif np.isnan(violation):
                raise np.linalg.LinAlgError()
            
            D = dl2_dx2 = hess(x)
            dl2_dx2_ = dl2_dx2.copy()
            dl2_dx2_[index] = dl_dx_
            dl_dx_[index] = l-target 
            
            dl2_dx2__inv = np.linalg.inv(dl2_dx2_)
            g = dl2_dx2__inv[:,index]
            v = np.dot(dl2_dx2__inv, dl_dx_)
            
            Dg = np.dot(D, g)
            vDg = np.dot(v, Dg)
            gDg = np.dot(g, Dg)
            vDv = np.dot(np.dot(v, D), v)
            
            p = 2 * (vDg-1) / gDg
            q = vDv / gDg
            
            _s = p*p/4 - q
            
            if _s >= 0:
                _s = np.sqrt(_s)
                s_ = - p/2
                s_1 = s_ + _s
                s_2 = s_ - _s
                
                v_sg_1 = v + s_1*g
                v_sg_2 = v + s_2*g
                measure1 = np.abs(np.dot(np.dot(v_sg_1, dl2_dx2_0), v_sg_1))
                measure2 = np.abs(np.dot(np.dot(v_sg_2, dl2_dx2_0), v_sg_2))
                
                if measure1 < measure2:
                    x -= v_sg_1
                else: 
                    x -= v_sg_2
            else:
                # TODO: do line search here! 
                x -= 0.1 * v
        else:
            l = fun(x)
            dl_dx_ = jac(x)
            violation = max(np.max(np.abs(dl_dx_[considered])), np.abs(l-target))
            return op.OptimizeResult(x=x, 
                                     fun=l,
                                     jac=dl_dx_,
                                     violation=violation,
                                     success=False, status=1,
                                     nfev=nmax+1, njev=nmax+1, nhev=nmax, 
                                     nit=nmax,
                                     x_track=np.array(x_track),
                                     message="Iteration limit exceeded"
                                     )
        
        return op.OptimizeResult(x=x, 
                                 fun=l,
                                 jac=jac(x),
                                 violation=violation,
                                 success=True, status=0,
                                 nfev=i, njev=i, nhev=i, 
                                 nit=i,
                                 x_track=np.array(x_track),
                                 message="Success"
                                 )
    except np.linalg.LinAlgError:
        return op.OptimizeResult(x=x0+np.nan, 
                                 fun=np.nan,
                                 jac=x0+np.nan,
                                 violation=np.nan,
                                 success=False, status=2,
                                 nfev=i, njev=i, nhev=i, 
                                 nit=i,
                                 x_track=np.array(x_track),
                                 message="Ill-conditioned matrix"
                                 )

def wald(x, hess, direction=None, alpha=ALPHA):
    for _ in [None]:
        diff = np.nan
        try:
            hinv = np.linalg.inv(hess)
        except Exception:
            message = "H is singular"
            break
        try:
            diff = np.sqrt(-np.diag(hinv) * chi2.ppf(alpha, df=1))
        except Exception:
            message = "H is not negative definite"
            break
    else: 
        message = "success"
        
    success = message == "success"
    if not direction:
        if not success:
            return None, None
        return x-diff, x+diff
    elif direction==1: 
        res = x+diff
    else: 
        res = x-diff
    return op.OptimizeResult(x=res, 
                             fun=np.nan,
                             success=success, 
                             nfev=0, njev=0, nhev=1, nit=0,
                             x_track=[],
                             f_track=[],
                             message=message
                             )


def root(index, fun, x0, alpha=ALPHA, method="df-sane"): #method='df-sane'):
    x2 = np.delete(x0, index)
    fun0 = fun(x0)
    target = fun0-chi2.ppf(alpha, df=1)/2
    def PL(xi):
        x = x2
        fun_nuiscance = lambda xx: -fun(np.insert(xx, index, xi))
        res = op.minimize(fun_nuiscance, x)
        x[:] = res.x
        print(xi, -res.fun-target)
        return -res.fun-target
    fullres = op.root(PL, x0[index], method=method)
    fullres.x = np.insert(x2, fullres.x, index)

def binsearch(index, direction, x0, fun, jac=None, alpha=ALPHA, initstep=1, 
              resulttol=1e-3, infstep=1e10, stepn=200, checkNo=5):
    fun = CounterFun(fun)
    
    if jac is not None:
        jac = CounterFun(jac)
        jac_nuiscance = lambda x: -np.delete(jac(np.insert(x, index, xi)), index)
    else: 
        jac_nuiscance = None
    
    xi = x0[index] + direction*initstep
    xi0 = x0[index]
    fun_nuiscance = lambda x: -fun(np.insert(x, index, xi))
    fun0 = fun(x0)
    target = fun0-chi2.ppf(alpha, df=1)/2
    xi_above = xi0
    xi_below = None
    x = np.delete(x0, index)
    f_trace = [fun0]
    x_trace = [x0]
    checkit = checkNo
    
    for i in range(stepn):
        opres = op.minimize(fun_nuiscance, x, jac=jac_nuiscance)
        f = -opres.fun
        print("{}: x_{}_d={:6.3f}, f_d={:6.3f}".format(i, index, xi-xi0, f-target))
        
        if np.isnan(f):
            f = -np.inf
        
        f_trace.append(f)
        x_trace.append(np.insert(opres.x, index, xi))
        
        if checkit < 0 and f > target:
            checkit = checkNo
            xi_below = None
            
        if f < target:
            xi_below = xi
            checkit = checkNo
        else:
            xi_above = xi
            x = opres.x
            checkit -= 1
            
        if xi_below is not None:
            if np.allclose(xi_above, xi_below, atol=resulttol):
                message = "success"
                success = True
                break
            if checkit == 0:
                xi = xi_below
                checkit -= 1
            else:
                xi = (xi_above+xi_below) / 2
        else:
            if np.abs(xi-xi0) > infstep and f > target:
                message = "unbounded"
                success = True
                break
            initstep *= 10
            xi += direction*initstep
    else:
        message = "iteration limit exceeded"
        success = False
    
    result = op.OptimizeResult(x=np.insert(x, index, xi), 
                             fun=f,
                             success=success, 
                             nfev=fun.evals, 
                             njev=(0 if jac is None else jac.evals), 
                             nhev=0,
                             nit=i,
                             x_track=np.array(x_trace),
                             f_track=np.array(f_trace),
                             message=message
                             )
    #print(result)
    return result

def gridsearch(index, direction, x0, fun, jac, hess, step=0.2, alpha=ALPHA,
               resulttol=1e-3, stepn=200, infstep=1000):
    
    f = fun0 = fun(x0)
    target = fun0-chi2.ppf(alpha, df=1)/2
    
    fun = CounterFun(fun)
    xi = xi0 = x0[index]
    
    dim = x0.size
    A = sparse.dok_matrix((dim, dim))
    A[index, index] = 1
    target_fun = lambda p: -fun(p)
    def target_fun(p): 
        res = -fun(p)
        #print(res+target, p[index]-xi) #, np.round(p,2)
        return res
    target_jac = CounterFun(lambda p: -jac(p))
    target_hess = CounterFun(lambda p: -hess(p))
    x = x0
    bound0 = np.full_like(x0, -np.inf)
    bound1 = np.full_like(x0, np.inf)
    for i in range(stepn+1):
        if i == stepn:
            x_prev = x
            f_prev = f
            xi += direction*infstep
        else:
            xi += direction*step
        #bounds = (-np.inf, xi) if direction < 0 else (xi, np.inf)
        bound0[index] = bound1[index] = xi
        bounds = (bound0, bound1)
        constraints = op.LinearConstraint(A, *bounds) 
        res = op.minimize(target_fun, x, jac=target_jac, hess=target_hess, constraints=constraints,
                             options=dict(maxiter=500), tol=resulttol, method="trust-constr") #, gtol=resulttol
        f = -res.fun
        print("{}: x_{}_d={:6.3f}, f_d={:6.3f}".format(i, index, xi-xi0, f-target))
        if f <= target:
            if step > resulttol:
                xi -= direction*step
                step /= 2
            else:
                message = "Success"
                #xi -= direction*step/2
                success = True
                break
        else:
            x = res.x
    else:
        if f < target:
            message = "iteration limit exceeded"
            success = False
            x = x_prev
            f = f_prev
        else:
            message = "Result is unbounded"
            success = True
    result = op.OptimizeResult(x=x, 
                             fun=f,
                             success=success, 
                             nfev=fun.evals, 
                             njev=target_jac.evals, 
                             nhev=target_hess.evals,
                             nit=i,
                             message=message
                             )
    return result
    

def bisection(index, direction, x0, fun_, jac=None, alpha=ALPHA, initstep=1, 
              resulttol=1e-3, constrainedMin=True, infstep=1e10, stepn=200):
    fun = CounterFun(fun_)
    xi = x0[index] + direction*initstep
    xi0 = x0[index]
    
    jac_nuiscance = None
    if constrainedMin:
        fun_nuiscance = lambda x: -fun(x)
        if jac is not None:
            jac = CounterFun(jac)
            unit = np.eye(1, x0.size, index).ravel()
            #def jac_nuiscance(x):
            #    print(x)
            #    jj = -jac(x)
            #    return jj
            #jac_nuiscance = CounterFun(jac_nuiscance)
            jac_nuiscance = CounterFun(lambda x: -jac(x))
        constraints = dict(type="eq", fun=lambda x: x[index]-xi, 
                           jac=lambda x: unit)
        dim = x0.size
        A = sparse.dok_matrix((dim, dim))
        A[index, index] = 1
    else:
        constraints = None
        fun_nuiscance = lambda x: -fun(np.insert(x, index, xi))
        if jac is not None:
            jac = CounterFun(jac)
            jac_nuiscance = lambda x: -np.delete(jac(np.insert(x, index, xi)), index)
    
    fun0 = fun(x0)
    target = fun0-chi2.ppf(alpha, df=1)/2
    f_above = fun0
    xi_above = xi0
    f_below = None 
    xi_below = None
    f_trace = [fun0]
    xi_trace = [xi0]
    x_trace_full = [x0]
    if not constrainedMin:
        x_trace = [np.delete(x0, index)]
    else:
        x_trace = x_trace_full
    
    for i in range(stepn):
        #if constrainedMin:
        #    bounds = (-np.inf, xi) if direction < 0 else (xi, np.inf)
        #    constraints = op.LinearConstraint(A, *bounds) 
        x = x_trace[np.argmin(np.abs(xi-np.array(xi_trace)))]
        x[np.isnan(x)] = x_trace[0][np.isnan(x)]
        for j in range(20):
            try:
                opres = op.minimize(fun_nuiscance, x, jac=jac_nuiscance, 
                                    constraints=constraints)
                if not np.isnan(opres.fun):
                    break
            except Exception:
                pass
            xi = (xi + xi_above) / 2
        else:
            message = "Error in function or jacobian"
            success = False
            break
            
        x_trace.append(opres.x)
        if not constrainedMin:
            x_trace_full.append(np.insert(opres.x, index, xi))
        f = -opres.fun
        f_trace.append(f)
        xi_trace.append(xi)
        print("{}: x_{}_d={:6.3f}, f_d={:6.3f}".format(i, index, xi-xi0, f-target))
        if np.isnan(f):
            print("Nan")
        if (xi_below is not None and (np.allclose(f, target, atol=resulttol) or
            np.allclose(xi_above, xi_below, atol=resulttol))):
            message = "success"
            success = True
            break
        elif np.abs(xi-xi0) > infstep and f > target:
            message = "unbounded"
            success = True
            break
        two_point = np.sum(np.array(f_trace)>target) + (f_below is not None) < 3
        if f >= target and f_below is not None:
            if f_below > target:
                if f < f_below:
                    f_above = f_below
                    xi_above = xi_below
                else:
                    two_point = True
                f_below = f
                xi_below = xi
            else:
                f_above = f
                xi_above = xi
        else:
            if f_below is not None and f_below > target:
                f_above = f_below
                xi_above = xi_below
            f_below = f
            xi_below = xi
            
        if two_point:
            p = (fun0-f)/(xi0-xi-0.5/xi0*(xi0**2-xi**2))
            a = -p/(2*xi0)
            q = fun0-p*(xi0)-a*xi0**2
        else:
            m = np.array([[xi0**2, xi0, 1],
                          [xi_above**2, xi_above, 1],
                          [xi_below**2, xi_below, 1]])
            a, p, q = np.dot(np.linalg.inv(m), [fun0, f_above, f_below])
        
            
        root = p**2/(4*a**2)-(q-target)/a
        if root < 0:
            initstep *= 10
            xi += direction*initstep
        else:
            dirFact = 1 if a<0 else -1
            xi = -p/(2*a) + dirFact * direction * np.sqrt(root)
            if np.abs(xi-xi0) > infstep:
                xi = direction*infstep*1.1
    else:
        message = "iteration limit exceeded"
        success = False
    result = op.OptimizeResult(x=x_trace_full[-1], 
                             fun=f,
                             success=success, 
                             nfev=fun.evals, 
                             njev=(0 if jac is None else jac.evals), 
                             nhev=0,
                             nit=i,
                             x_track=np.array(x_trace_full),
                             f_track=np.array(f_trace),
                             message=message
                             )
    print(result.x, result.fun)
    return result

def constrained_max(index, direction, x0, fun, jac=None, alpha=ALPHA, nit=500):
    fun0 = fun(x0)
    print("start", index, direction)
    
    target = fun0-chi2.ppf(alpha, df=1)/2
    preconditioning = 0.001
    target_fun = lambda x: -direction*x[index]
    unit = -np.eye(1, x0.size, index).ravel()*direction
    target_jac = lambda x: unit
    constraints = dict(type="eq", fun=lambda x: (fun(x)-target)*preconditioning, 
                       jac=lambda x: jac(x)*preconditioning)
    
    try:
        result = op.minimize(target_fun, x0, jac=target_jac, constraints=constraints,
                             options=dict(maxiter=500))
    except Exception:
        return op.OptimizeResult(fun=np.nan,
                                 success=False,
                                 x=x0,
                                 nit=0)
    print("done", index, direction)
    return result
    
    
def mixed_min(index, direction, x0, fun, jac=None, alpha=ALPHA):
    
    preconditioning = 0.01
    fun0 = fun(x0)
    target = fun0-chi2.ppf(alpha, df=1)/2
    f = lambda x: (2*np.square(fun(x)-target) - direction*x[index])*preconditioning
    
    
    if jac is not None:
        unit = np.eye(1, x0.size, index).ravel()
        j = lambda x: (4*jac(x)*(fun(x)-target) - direction*unit)*preconditioning
    else:
        j = None
    #j = Gradient(f, step=1e-7)
    result = op.minimize(f, x0, jac=j) #, method='trust-exact')
    print("index {:2d}: f_diff={:7.3f} (direction {})".format(index, fun(result.x)-target, direction))
    result.fun = fun(result.x)
    
    print(target-result.fun)
    #print(result)
    return result

def fixedFun(f, origArg, flex):
    def fun(x):
        xx = origArg.copy().astype(x.dtype)
        xx[flex] = x
        return f(xx)
        
    if isinstance(f, CounterFun):
        fun.evals = lambda: f.evals
    return fun

class BaseTester():
    def __init__(self):
        self.ML = None
        self._MLE = None
        self.x = None
        self.dataN = None
        self.data = None
        self.convert = False
        self.convertIndices = None
        self.fixed = None
        self.dim = None
        self.test_prediction = False
    
    @property
    def MLE(self):
        if self.fixed is None:
            return self._MLE
        return np.delete(self._MLE, self.fixed)
        
    @MLE.setter
    def MLE(self, val):
        self._MLE = val
        
    def convertPos(self, x):
        return convertPos(x, self.convertIndices, self.convert) 
    def convertPosInv(self, x):
        return convertPosInv(x, self.convertIndices, self.convert) 
    
    @property
    def funs(self):
        return self.get_funs()
        
    def get_funs(self):
        raise NotImplementedError("get_funs has not been implemented yet")


class DynamicalSystemTester(BaseTester):
    def __init__(self, *sim_args, **sim_kwargs):
        
        BaseTester.__init__(self)
        
        self.convert = CONVERT
        self.dim = 8
        self.simulate(*sim_args, **sim_kwargs)
    
    @staticmethod
    def rss_dyn(f, t, y, dim):
        tmax = np.max(t)
        def ff(p):
            fp = lambda t, x: f(t, x, p[dim:])
            try:
                res = np.sum(np.square(
                    integrate.solve_ivp(fp, (0, tmax), p[:dim], rtol=1e-9, atol=1e-11, dense_output=True, vectorized=True).sol(t) - y))
            except Exception:
                return 1000000
            #print(res)
            return res
        return ff
    
    @staticmethod
    def f_constr(f, t, y, dim, fact, target, tol):
        tmax = np.max(t)
        tt = np.linspace(0, tmax, 1000)
        def ff(p):
            fp = lambda t, x: f(t, x, p[dim:-1])
            try:
                sol = integrate.solve_ivp(fp, (0, tmax), p[:dim], rtol=1e-9, atol=1e-11, dense_output=True, vectorized=True)
                res = fact*np.sum(np.square(sol.sol(t) - y)) + (p[-1]-np.mean(sol.sol(tt)[0]))**2/tol**2*target
            except Exception:
                res = 1000000
            #print(res)
            return res
        return ff
    
    @staticmethod
    def f_prime(t, x, params):
        r, K, a, h, c, d = params
        x, y = x
        conversion = a*y*x/(1+h*x)
        return np.array((r*x*(1-x/K) - conversion, 
                         (c*conversion-y*d)))
    
    def get_rss(self):
        rss_ = DynamicalSystemTester.rss_dyn(DynamicalSystemTester.f_prime, *self.data, 2)
        return lambda p: rss_(convertPos(p))
    
    def get_funs(self):
        rss = self.get_rss()  
        fact = self.fact
        f_ = lambda x: fact*rss(x)
        if self.fixed is not None:
            MLE = self._MLE
            flex = np.ones_like(MLE, dtype=bool)
            flex[self.fixed] = False
            f_ = fixedFun(f_, MLE, flex)
        f = CounterFun(f_)
        g = Gradient(f, step=1e-7, method='complex')
        h = Hessian(f, step=1e-7, method='complex')
        return f, g, h
    
    def get_funs_constr(self):
        varMLE = self.ML / (self.dataN*2)
        fact = -0.5/varMLE
        dim = 2
        f = DynamicalSystemTester.f_constr(DynamicalSystemTester.f_prime, 
                                           *self.data, dim, fact, 
                                           chi2.ppf(.95, df=1)/2, 1e-3)
        g = Gradient(f, step=1e-8, method='complex')
        h = Hessian(f, step=1e-8, method='complex')
        t = self.data[0]
        tmax = np.max(t)
        p = self.MLE
        sol = integrate.solve_ivp(DynamicalSystemTester.f_prime, (0, tmax), p[:dim], rtol=1e-9, atol=1e-11, dense_output=True, vectorized=True)
        tt = np.linspace(0, tmax, 1000)
        self.cMLE = np.append(self.MLE, np.mean(sol.sol(tt)[0]))
        
        return f, g, h
        
        
    def simulate(self, params=None, dataN=50, tEnd=20, std=3, plot=True):
        
        if params is None:
            params = np.array((50, 25, 2, 200, 0.1, 0.01, 0.3, 1))
            
        print("simulating data")
        
        t = np.linspace(0, tEnd, dataN)
        paramsFull = params
        x0, params = params[:2], params[2:]
        res = integrate.solve_ivp(partial(self.f_prime, params=params), 
                                  (0, tEnd), x0, dense_output=True, 
                                  vectorized=True)
        x = np.maximum(res.sol(t) + np.random.randn(*t.shape)*std, 0)
        self.data = [t, x]
        #print(x[:,-1])
        rss = self.get_rss()   
        print("creating likelihood function")
        def rsss(p):
            #print(convertPos(p))
            res = rss(p)
            print(res)
            return res
        self.params = convertPosInv(paramsFull)
        
        print(rss(self.params))
        g = Gradient(rss, method='complex') #Gradient(rss, 1e-6, num_steps=4)
        h = Hessian(rss, method='complex')
        fit = op.minimize(rsss, self.params, jac=g, hess=h) #, options=dict(maxiter=1))
        print(fit)
        xMLE, paramsMLE = fit.x[:2], fit.x[2:]
        
        varMLE = fit.fun / dataN**2
        self.fact = -0.5/varMLE
        
        self.ML = fit.fun * self.fact
        self.MLE = fit.x
        self.x = paramsFull
        self.dataN = dataN
        if plot:
            print("plotting")
            tt = np.linspace(0, tEnd, tEnd*100)
            resHat = integrate.solve_ivp(partial(self.f_prime, params=convertPos(paramsMLE)), 
                                         (0, tEnd), convertPos(xMLE), dense_output=True, 
                                         vectorized=True)
            for y in res.sol(tt):
                plt.plot(tt, y)
            for y in resHat.sol(tt):
                plt.plot(tt, y)
            for y in x:
                plt.plot(t, y, '.')
            plt.show()
        
        
class LogRegressTester(BaseTester):
    def __init__(self, *sim_args, seed=None, **sim_kwargs):
        BaseTester.__init__(self)
        np.random.seed(seed)
        self.simulate(*sim_args, **sim_kwargs)
    
    def get_funs(self):
        covariates, switch = self.data
        if self.powers:
            covariateN = self.covariateN
            #f = lambda p: (-ag.sum(ag.log(1+ag.exp(-ag.sum(p[:covariateN]*covariates[switch:]**p[covariateN:-1], 1)-p[-1])))
            #               -ag.sum(ag.log(1+ag.exp(ag.sum(p[:covariateN]*covariates[:switch]**p[covariateN:-1], 1)+p[-1]))))
            f_ = lambda p: (-ag.sum(ag.log(1+ag.exp(-ag.sum(p[:covariateN]*ag.power(covariates[switch:],p[covariateN:-1]), 1)-p[-1])))
                           -ag.sum(ag.log(1+ag.exp(ag.sum(p[:covariateN]*ag.power(covariates[:switch],p[covariateN:-1]), 1)+p[-1]))))
            def f_(p): 
                p = self.convertPos(p)
                return (-ag.sum(ag.log(1+ag.exp(-ag.sum(p[:covariateN]*ag.power(covariates[switch:],p[covariateN:-1]), 1)-p[-1])))
                        -ag.sum(ag.log(1+ag.exp(ag.sum(p[:covariateN]*ag.power(covariates[:switch],p[covariateN:-1]), 1)+p[-1]))))
        else:
            f_ = lambda p: (-ag.sum(ag.log(1+ag.exp(-ag.sum(covariates[switch:]*p[:-1], 1)-p[-1])))
                           -ag.sum(ag.log(1+ag.exp(ag.sum(covariates[:switch]*p[:-1], 1)+p[-1]))))
        
        
        if self.test_prediction:
            testData = self.testData
            tol = 0.001
            target = 2
            f = lambda p: (f_(p[:-1]) - 
               (p[-1]-np.mean(1/(1+np.exp(
                   -np.sum(p[:covariateN]*testData**p[covariateN:-2], 1)
                   -p[-2]))))**2/tol**2*target)
            def f(p): 
                diff = (p[-1]-np.mean(1/(1+np.exp(
                   -np.sum(p[:covariateN]*testData**p[covariateN:-2], 1)
                   -p[-2]))))
                res = f_(p[:-1]) - (diff/tol)**2*target
                if isinstance(diff, float):
                    #print(diff)
                    pass
                return res
        else:
            f = f_
            
        f = CounterFun(f)
        
        if self.fixed is not None:
            MLE = self._MLE
            flex = np.ones_like(MLE, dtype=bool)
            flex[self.fixed] = False
            f = fixedFun(f, MLE, flex)
            g = Gradient(f, method='complex', step=1e-7)
            h = Hessian(f, method='complex', step=1e-7)
        else:
            #g = AgGradient(f)
            #h = AgHessian(f)
            g = Gradient(f, method='complex', step=1e-7)
            h = Hessian(f, method='complex', step=1e-7, num_steps=1)
        return f, g, h
    
    def simulate(self, params=None, dataN=1000, testDataN=0, powers=True, mode="3"):
        
        print("simulating data")
        
        if params is None:
            if mode=="3":
                params = np.array([5, 0.5, -10])
            elif mode=="11":
                params = np.array([5, 2, -1, -3, -2, 0.2, 1, 0.1, 0.2, 0.5, -1])
            elif mode=="11cx":
                params = np.array([0.8, 0.2, -0.6, -1, -1, 0.2, 0.5, 0.1, -0.2, 0.2, 2])
                powers = False
        if powers:
            covariateN = (len(params)-1)//2
        else:
            covariateN = len(params)-1
        
        self.dataN = dataN
        self.dim = len(params)
        if powers:
            self.convert = True
            self.convertIndices = slice((self.dim-1)//2, -1)
        allDataN = testDataN + dataN
        
        covariates = np.zeros((allDataN, covariateN), dtype=int)
        m = 5
        p = 0.5
        p2 = 0.2
        for i in range(covariateN):
            if not i % 2:
                covariates[:,i] = np.random.negative_binomial(m, p, allDataN)
            else:
                covariates[:,i] = np.random.binomial(covariates[:,i-1], p2, allDataN) #+ np.random.poisson(0.5, allDataN)
        
        covariates = covariates+1e-16
        
        self.testData = covariates[dataN:]
        covariates = covariates[:dataN]
        
        if powers:
            data = np.random.binomial(1, 1/(1+np.exp(-np.sum(params[:covariateN]*covariates**params[covariateN:-1], 1)-params[-1])))
        else:
            data = np.random.binomial(1, 1/(1+np.exp(-np.sum(covariates*params[:-1], 1)-params[-1])))
        print(data.mean())
        if powers:
            print(np.round(1/(1+np.exp(-np.sum(params[:covariateN]*covariates**params[covariateN:-1], 1)-params[-1])),2))
        else:
            print(np.round(1/(1+np.exp(-np.sum(params[:covariateN]*covariates, 1)-params[-1])),2))
        self.covariateN = covariateN
        self.powers = powers
        
        order = np.argsort(data)
        covariates = covariates[order]
        data = data[order]
        
        switch = np.argmax(data)
        
        self.data = [covariates, switch]
        
        f, j, h = self.funs
        
        print(f(self.convertPosInv(params)))
        ff = lambda p: -f(p)
        jj = lambda p: -j(p)
        hh = lambda p: -h(p)
        fit = op.minimize(ff, self.convertPosInv(params), jac=jj)
        fit = op.minimize(ff, fit.x, jac=jj, hess=hh, method='trust-exact')
        print(j(params))
        print(fit)
        print(params)
        
        self.test_prediction = testDataN
        if self.test_prediction:
            testP = 1/(1+np.exp(-np.sum(params[:covariateN]*self.testData**params[covariateN:-1], 1)-params[-1]))
            testPMLE = 1/(1+np.exp(-np.sum(fit.x[:covariateN]*self.testData**fit.x[covariateN:-1], 1)-fit.x[-1]))
            self.x = np.append(params, np.mean(testP))
            self.MLE = np.append(fit.x, np.mean(testPMLE))
        else:
            self.x = params
            self.MLE = fit.x
        self.ML = -fit.fun
    
    
    def simulate_(self, params=None, dataN=1000, powers=True):
        
        if params is None:
            params = 5
        if isinstance(params, int):
            params = np.random.randn(params)
            
        self.dim = len(params)
        if powers:
            covariateN = (len(params)-1)//2
        else:
            covariateN = len(params)-1
        if powers:
            params[covariateN:-1] = np.random.rand(covariateN)
        
        print("simulating data")
        cov = np.random.randn(covariateN, covariateN)+10
        cov = np.dot(cov, cov.T)
        
        
        m = np.random.randn(covariateN)
        
        covariates = np.random.multivariate_normal(m, cov, dataN)/50
        #covariates[:,1] = covariates[:,0] + np.random.rand(dataN)*0.01
        
        if powers:
            shift = 0.01 - np.min(covariates)
            params[-1] -= shift
            covariates += shift
            data = np.random.binomial(1, 1/(1+np.exp(-np.sum(params[:covariateN]*covariates**params[covariateN:-1], 1)-params[-1])))
        else:
            data = np.random.binomial(1, 1/(1+np.exp(-np.sum(covariates*params[:-1], 1)-params[-1])))
        
        self.covariateN = covariateN
        self.powers = powers
        
        order = np.argsort(data)
        covariates = covariates[order]
        data = data[order]
        
        switch = np.argmax(data)
        
        self.data = [covariates, switch]
        
        f, j, h = self.funs
        
        print(f(params))
        ff = lambda p: -f(p)
        
        self.x = params
        print(j(params))
        fit = op.minimize(ff, params, jac=j, hess=h, method='trust-exact')
        print(fit)
        print(params)
        
        self.ML = -fit.fun
        self.MLE = fit.x

class H14Tester(BaseTester):
    
    def __init__(self, fixed=[0], vals=[25]): #fixed=[0, 1, 2], vals=[25, 0, 0]):
        BaseTester.__init__(self)
        self.convert = CONVERT
        self.dim = 7-len(fixed)
        
        r_source = robjects.r['source']
        print("Executing R script")
        robjects.r('setwd("test_CI/Histone_H1")')
        r_source("get_f.R")
        #x0 = np.array([0.02, 0., 0, 0.090083433, 0.015029212, 0.001207238, 0.002303329])
        x0 = np.array([0, 0.03429, 0, 0.37168, 0.05365, 0, 0.00794])
        x0 = np.array([0, 0.04024, 0.00794, 0.36572, 0.05365, 0, 0 ])
        x0 = np.array([0, 0, 0.00794, 0.40596, 0.04912, 0.00453, 0 ])
        x0 = np.array([0, 0, 0, 4.05836e-01, 4.91267e-02, 3.79864e-03, 8.67439e-03])
        #x0 = np.array([0] + [1e-3]*6)
        if fixed is not None:
            fixed = np.array(fixed, dtype=int)
            fixed.sort()
            insertIndices = fixed - np.arange(fixed.size)
            x0[fixed] = vals
        x0 = convertPosInv(x0)
        ff = robjects.r['f']
        vals = x0[fixed]
        x0 = np.delete(x0, fixed)
        def f(x):
            if fixed is not None and len(fixed):
                x = np.insert(x, insertIndices, vals)
            try:
                return -ff(robjects.FloatVector(convertPos(x)))[0]
            except Exception: #rinterface.RRuntimeError:
                return -np.inf
        
        fun2 = lambda x: -f(x)
        
        j = Gradient(fun2, 1e-5, num_steps=2)
        h = Hessian(fun2, 1e-5, num_steps=2)
        
        opres = op.minimize(fun2, x0, jac=j, hess=h, method='trust-exact')
        print(opres)
        n = robjects.r['dataN'][0]
        var = opres.fun / n
        fact = 0.5/var
        self.constVals = (insertIndices, vals)
        
        self.MLE = opres.x
        self.ML = -fact * opres.fun 
        self.fact = fact
        self.x = convertPos(opres.x)
        
        print("done", self.x, self.ML, self.fact)
        
    def get_funs(self):
        
        insertIndices, vals = self.constVals
        from rpy2 import robjects
        r_source = robjects.r['source']
        print("Executing R script")
        #robjects.r('setwd("test_CI/Histone_H1")')
        r_source("get_f.R")
        ff = robjects.r['f']
        
        def f(x):
            if insertIndices is not None and len(insertIndices):
                x = np.insert(x, insertIndices, vals)
            try:
                return -ff(robjects.FloatVector(convertPos(x)))[0]
            except Exception: #rinterface.RRuntimeError:
                return -np.inf
        
        fact = self.fact
        
        fun = CounterFun(lambda x: fact*f(x))
        
        if self.fixed is not None:
            MLE = self._MLE
            flex2 = np.ones_like(MLE, dtype=bool)
            flex2[self.fixed] = False
            fun = fixedFun(fun, MLE, flex2)
        
        j = Gradient(fun, 1e-5, num_steps=2)
        h = Hessian(fun, 1e-5, num_steps=2)
    
        return fun, j, h        
        
def find_CI(tester, index, direction, method="RVM", f_track=False):
    nmax = 600
    infstep = 1e3
    resulttol=1e-5
    if method == "Wald":
        tmpPos = tester.convertPos
        tester.convertPos = False
    f_, g, h = tester.funs
    if method == "Wald":
        tester.convertPos = tmpPos 
    
    if f_track:
        f = TrackFun(f_)
    else:
        f = f_
    try:
        if method == "RVM":
            res = find_CI_bound(tester.MLE, f, index, direction, 
                                g, h, nmax=nmax, apprxtol=0.5,
                                alpha=ALPHA, disp=False,
                                infstep=infstep)
        elif method == "RVM_psI":
            res = find_profile_CI_bound_mpinv(index, direction, tester.MLE,
                                         f, g, h, nmax=nmax, apprxtol=0.5,
                                         alpha=ALPHA, disp=False, infstep=infstep)
        elif method == "VM":
            res = venzon_moolgavkar(index, direction, tester.MLE,
                                         f, g, h, nmax=nmax,
                                         alpha=ALPHA, disp=False)
        elif method == "bisection":
            res = bisection(index, direction, tester.MLE, f, g, stepn=nmax, 
                            infstep=infstep, resulttol=resulttol)
        elif method == "binsearch":
            res = binsearch(index, direction, tester.MLE, f, g, stepn=nmax, 
                            infstep=infstep, resulttol=resulttol)
        elif method == "gridsearch":
            res = gridsearch(index, direction, tester.MLE, f, g, h, stepn=nmax,
                             resulttol=resulttol)
        elif method == "Wald":
            res = wald(tester.MLE, h(tester.convertPos(tester.MLE)), direction)
            return res, 0
        elif method == "mixed_min":
            res = mixed_min(index, direction, tester.MLE, f, g)
        elif method == "constrained_max":
            res = constrained_max(index, direction, tester.MLE, f, g)
        else:
            raise ValueError("Method {} unknown.".format(method))
        if f_track:
            return res, f.track
        return res, f_.evals
    except Exception:
        return op.OptimizeResult(x=tester.MLE*np.nan, 
                             fun=np.nan,
                             success=False, 
                             nfev=f_.evals, 
                             njev=0, 
                             nhev=0,
                             nit=0,
                             x_track=np.array(tester.MLE),
                             f_track=np.array(tester.ML),
                             message="Error when computing the CI."
                             ), f_.evals

def find_CIs(tester, method="RVM", indices=None, printCI=True, 
             multiprocessing=True):
    print("Using method", method)
    if indices is None:
        n = tester.dim
        indices = range(n)
    else:
        n = len(indices)
    
    #results = list(starmap(find_CI, zip(repeat(tester), list(range(4, dim))*2, [1]*dim + [-1]*dim, repeat(method))))
    args = zip(repeat(tester), list(indices)*2, [-1]*n + [1]*n, repeat(method))
    #args = zip([tester], [10], [1], repeat(method))
    
    
    if multiprocessing:
        with Pool() as pool:
            results = pool.starmap(find_CI, args)
    else:
        results = starmap(find_CI, args)
    
    
    results = list(results)
    
    results2 = []
    
    target = tester.ML-chi2.ppf(ALPHA, df=1)/2
    tol = 1e-3
    if printCI:
        print("============== {:<15}================".format(method))
    for i, j in enumerate(indices):
        result1, feval1 = results[i]
        result2, feval2 = results[i+n]
        left = tester.convertPos(result1.x)[j]
        right = tester.convertPos(result2.x)[j]
        leftOriginal = result1.x[j]
        rightOriginal = result2.x[j]
        if printCI:
            print("[{:14.5f}{}, {:9.5f}, {:14.5f}{}], ({:4d}, {:4d}), ({:6d}, {:6d})".format(
                left, "*" if not result1.success else " ",
                tester.x[j], right, 
                "*" if not result2.success else " ", result1.nit, result2.nit,
                feval1, feval2))
            
            if method=="RVM" and (result1.fun < target-tol or result2.fun < target-tol):
                print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
                print(result1.fun -(target-tol), result2.fun - (target-tol))
                print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
            
        results2.append([left, right, leftOriginal, rightOriginal, result1.fun >= target-tol, 
                         result2.fun >= target-tol, result1.success, 
                         result2.success, feval1, feval2])
    if printCI:
        print("==============================================")
        print()
            
    return results2

def test_dynamical_system():
    #                      r, K,   a,    h,     c,  d
    p = np.array((50, 25, 2, 200, 0.1, 0.1, 0.3, 1))
    tester = DynamicalSystemTester(None) #p, std=3)
    find_CIs(tester, "mixed_min")
    find_CIs(tester, "RVM")
    find_CIs(tester, "constrained_max")
    find_CIs(tester, "Wald")
    find_CIs(tester, "binsearch")
    find_CIs(tester, "bisection")
    return

def test_LogRegress(methods):
    tester = LogRegressTester(dataN=500)
    #find_CI(tester, 3, 1, "constrained_max")
    for m in methods:
        find_CIs(tester, m)

def test_LogRegress_pred():
    tester = LogRegressTester(testDataN=10)
    indices = [tester.dim]
    find_CIs(tester, "constrained_max", indices)
    find_CIs(tester, "mixed_min", indices)
    find_CIs(tester, "Wald", indices)
    find_CIs(tester, "RVM", indices)
    find_CIs(tester, "binsearch", indices)
    find_CIs(tester, "bisection", indices)
    return

def test_H14():
    tester = H14Tester()
#     find_CIs(tester, "RVM")
#     find_CIs(tester, "RVM_psI")
#     find_CIs(tester, "constrained_max")
#     find_CIs(tester, "mixed_min")
#     find_CIs(tester, "binsearch")
#     find_CIs(tester, "bisection")
#     find_CIs(tester, "gridsearch")
#     find_CIs(tester, "Wald")
    find_CIs(tester, "VM")
    return

def create_plot(tester, considered, additional=[], methods=["RVM"], fileName=None):
    
    print("Creating figure for", tester, considered, additional)
    
    fixed = list(range(tester.dim))
    for i in considered:
        fixed.remove(i)
    for i in additional:
        fixed.remove(i)
    
    if methods == ["RVM"]:
        rvmplot = True
    
    consIndices = np.sort(np.concatenate((considered, additional)))
    index = np.searchsorted(consIndices, considered[0])
    ind2 = np.searchsorted(consIndices, considered[1])
    plotinds = np.array([index, ind2])
    flex = []
    for i in additional:
        flex.append(np.searchsorted(consIndices, i))
    flex = np.array(flex)
    
    tester.fixed = fixed
    fun = tester.funs[0]
    
    xmax = -np.inf
    xmin = np.inf
    ymax = -np.inf
    ymin = np.inf
    x, y = [], []
    x2, y2 = [], []
    strings = []
    for method in methods:
        r1, track1 = find_CI(tester, index, -1, method, f_track=True)
        r2, track2 = find_CI(tester, index, 1, method, f_track=True)
        xmax = max(xmax, tester.convertPos(r2.x)[considered[0]], tester.convertPos(r1.x)[considered[0]])
        xmin = min(xmin, tester.convertPos(r2.x)[considered[0]], tester.convertPos(r1.x)[considered[0]])
        ymax = max(ymax, tester.convertPos(r2.x)[considered[1]], tester.convertPos(r1.x)[considered[1]])
        ymin = min(ymin, tester.convertPos(r2.x)[considered[1]], tester.convertPos(r1.x)[considered[1]])
        strings.append(method + ": [{:6.3f}, {:6.3f}]".format(r1.x[considered[0]], r2.x[considered[0]]))
        xx, yy = tester.convertPos(np.concatenate((track1[::-1], track2)).T)[considered]
        x.append(xx)
        y.append(yy)
        if rvmplot:
            xx2, yy2 = tester.convertPos(np.concatenate((r1.x_track[::-1], r2.x_track)).T)[considered]
            x2.append(xx2)
            y2.append(yy2)
    for s in strings:
        print(s)
    
    if rvmplot:
        xmax = min(np.max(x[0]), 99)
        xmin = np.min(x[0])
        ymax = min(np.max(y[0]), 99)
        ymin = np.min(y[0])
        
        xOrder = 10**np.floor(np.log10(xmax))/2
        yOrder = 10**np.floor(np.log10(ymax))
        xmax = (xmax // xOrder + 1) *xOrder
        xmin = max((xmin // xOrder) * xOrder, -10)
        ymax = (ymax // yOrder + 1) *yOrder
        ymin = (ymin // yOrder) * yOrder
        
        if tester.dataN >= 10000:
            xmin, xmax = 0, 11
        
    else:
        xdiff = xmax-xmin
        ydiff = ymax-ymin
        
        xmax += 0.1*xdiff
        xmin -= 0.1*xdiff
        ymax += 0.1*ydiff
        ymin -= 0.1*ydiff
    #print(np.arctan(r1.x[index])/np.pi+0.5, np.arctan(r2.x[index])/np.pi+0.5)
    
    if not (r1.success and r2.success):
        #raise Exception("no success")
        pass
    
    #x, y = tester.convertPos(np.concatenate((r1.x_track[::-1], r2.x_track))).T[considered]
    """
    xmax = np.max(x[0])
    xmin = np.min(x[0])
    ymax = np.max(y[0])
    ymin = np.min(y[0])
    """
    #xmin, xmax = -100, 2
    #ymin, ymax = -2, 2
    #ymax = np.mean(y)
    #xmax = np.mean(x)
    
    
    #xmin = max(0, xmin)
    #ymin = max(0, ymin)
    plotn = 200
    XX = np.linspace(xmin, xmax, plotn)
    YY = np.linspace(ymin, ymax, plotn)
    X, Y = np.meshgrid(XX, YY)
    
    Z = np.zeros_like(X)
    
    x0 = tester.MLE
    target = tester.ML-chi2.ppf(ALPHA, df=1)/2
    print(target, tester.ML)
    figsize = (3.5, 3)
    rect = [0.15, 0.18, 0.75, 0.78]
    plt.figure(figsize=figsize).add_axes(rect)
    if rvmplot:
        for i, xx2 in enumerate(XX):
            print(i)
            for j, yy2 in enumerate(YY):
                if tester.convert and (xx2<0 or yy2<0):
                    Z[j, i] = np.nan 
                else:
                    if additional:
                        x02 = x0.copy()
                        x02 = tester.convertPos(x02)
                        x02[index] = xx2
                        x02[ind2] = yy2
                        x02 = tester.convertPosInv(x02)
                        #x02[flex] = -x02[ind2]
                        ff = fixedFun(lambda x: -fun(x), x02, flex)
                        #x03 = op.basinhopping(ff, x02[flex], 10).x
                        if len(additional) == 1:
                            Z[j, i] = -op.minimize_scalar(ff, bounds=(-300, 100), method="bounded").fun
                        else:
                            Z[j, i] = -op.minimize(ff, x02).fun
                        #Z[j, i] = -op.minimize(ff, x02[flex]).fun
                    else:
                        Z[j, i] = fun(tester.convertPosInv(np.array([xx2, yy2])))
        Z[np.isnan(Z)] = -np.inf
        plt.pcolormesh(X, Y, np.maximum(Z, tester.ML-50), shading='gouraud') #cmap=cmap) #, 
    #"""
    #plt.plot(xx, yy, marker = 'x', color='C5')
    #plt.plot(x, y, marker = 'o', color='C1')
    markers = ["o", "v", "s", "h", "D"]
    colors = ["C1", "C0", "C2", "C3", "C4", "C5"]
    plt.contour(X, Y, Z, levels=[target]) #cmap=cmap) #, 
    if rvmplot:
        for xx, yy, marker, color, method in zip(x2, y2, markers, colors, methods):
            plt.plot(xx, yy, marker=marker, color=color)
        
        
    for xx, yy, marker, color, method in zip(x, y, markers, colors, methods):
        opacity = 1 
        if not rvmplot:
            if method=="RVM": 
                color = 'k'
            else: 
                opacity = 0.5
            markersize = 5
        else:
            markersize = 3
        plt.plot(xx, yy, marker=marker, color=color, alpha=opacity, linewidth=0.3, 
                 label=method, markersize=markersize)
            
    plt.plot(*tester.convertPos(x0)[plotinds], marker = 'o', color="C3", label="MLE", linewidth=0)
    
    plt.xlim(np.min(X), np.max(X))
    plt.ylim(np.min(Y), np.max(Y))
    if rvmplot:
        plt.xlabel(r"$\beta_1$")
        plt.ylabel(r"$\alpha_1$")
        plt.locator_params(nticks=3, nbins=3)
    else:
        plt.xlabel(r"$x_{}$".format(considered[0]))
        plt.ylabel(r"$x_{}$".format(considered[1]))
        plt.legend()
    
    #print(xx)
    #print(yy)
    
    #ax.plot_surface(X, Y, Z, rstride=1, cstride=1,
    #                   linewidth=0, antialiased=True)
    if rvmplot:
        fileName = "rvm" + str(tester.dataN)
    if fileName:
        plt.savefig(fileName+".png", dpi=1000)
        plt.savefig(fileName+".pdf", dpi=1000)
    if not fileName or rvmplot:
        plt.show()

def benchmark_(methods, simkwargs):
    tester = LogRegressTester(**simkwargs)
    return [find_CIs(tester, method, printCI=True,
                     multiprocessing=False) for method in methods]

def benchmark(methods, nsim=100, **simkwargs):
    print("Benchmarking {} realizations.".format(nsim))
    print("simargs:", simkwargs)
    
    dtype = [("left", float), ("right", float), ("leftOriginal", float), 
             ("rightOriginal", float), ("leftAdmissible", bool), 
             ("rightAdmissible", bool), ("leftSuccess", bool), 
             ("rightSuccess", bool), ("leftEvals", int), ("rightEvals", int)]
    
    stats = defaultdict(lambda: dict(success=[], error=[], evals=[], 
                                     errorTot=[], evalsTot=[],
                                     errorReduced=[], largeErrors=[],
                                     largeErrorsTot=[]))
    
    bound = 1000
    tol = 1e-3
    rtol = 0.05
    with Pool() as pool:
        for results_ in pool.starmap(benchmark_, zip(repeat(methods), repeat(simkwargs, nsim)),
                                     chunksize=1):
            results = np.core.records.fromarrays(np.array(results_).T, dtype=dtype)
            
            lefts = np.ma.array(results["left"], mask=~results["leftAdmissible"])
            rights = np.ma.array(results["right"], mask=~results["rightAdmissible"])
            
            results["left"] = np.maximum(results["left"], -bound)
            results["right"] = np.minimum(results["right"], bound)
            
            
            trueLeft = np.min(lefts, 1)
            trueLeft[trueLeft.mask] = np.nan
            trueLeft = trueLeft.data
            trueRight = np.max(rights, 1)
            trueRight[trueRight.mask] = np.nan
            trueRight = trueRight.data
            
            for method, result in zip(methods, results.T):
                methodStats = stats[method]
                methodStats["success"].extend(
                    (result["leftSuccess"] & np.isclose(result["left"], trueLeft, rtol=rtol, atol=tol))
                     | ((result["leftOriginal"] <= -bound) & result["leftAdmissible"]))
                methodStats["success"].extend(
                    (result["rightSuccess"] & np.isclose(result["right"], trueRight, rtol=rtol, atol=tol))
                     | ((result["rightOriginal"] >= bound) & result["rightAdmissible"]))
                leftError = np.abs(result["left"]-trueLeft)
                rightError = np.abs(np.abs(result["right"]-trueRight))
                methodStats["error"].extend(leftError[result["leftSuccess"]])
                methodStats["error"].extend(rightError[result["rightSuccess"]])
                methodStats["errorTot"].extend(leftError[trueLeft > -bound])
                methodStats["errorTot"].extend(rightError[trueRight < bound])
                largeLeftError = ((leftError > 10) & result["leftSuccess"]) | ((trueLeft <= -bound) & (result["left"] > 10-bound))
                largeRightError = ((rightError > 10) & result["rightSuccess"]) | ((trueRight >= bound) & (result["right"] < bound-10))
                largeLeftErrorTot = (leftError > 10) & ~((trueLeft <= -bound) & (result["left"] < 10-bound))
                largeRightErrorTot = (rightError > 10) & ~((trueRight >= bound) & (result["right"] > bound-10))
                if method=='RVM' and (largeLeftError.any() or largeRightError.any()):
                    print("####################################")
                    print(results)
                    print("####################################")
                methodStats["errorReduced"].extend(leftError[result["leftSuccess"] & (~largeLeftError)])
                methodStats["errorReduced"].extend(rightError[result["rightSuccess"] & (~largeRightError)])
                methodStats["largeErrors"].extend(largeLeftError)
                methodStats["largeErrors"].extend(largeRightError)
                methodStats["largeErrorsTot"].extend(largeLeftErrorTot)
                methodStats["largeErrorsTot"].extend(largeRightErrorTot)
                
                #methodStats["errorTot"].extend(np.abs(result["left"]-trueLeft)[result["leftOriginal"] > -bound])
                #methodStats["errorTot"].extend(np.abs(result["right"]-trueRight)[result["rightOriginal"] < bound])
                methodStats["evals"].extend(result["leftEvals"][result["leftSuccess"]])
                methodStats["evals"].extend(result["rightEvals"][result["rightSuccess"]])
                methodStats["evalsTot"].extend(result["leftEvals"])
                methodStats["evalsTot"].extend(result["rightEvals"])
        
    for method in methods:
        methodStats = stats[method]
        print("{:<15}: success={:7.5f}, error={:8.4f}, errorReduced={:8.4f}, errorTot={:8.4f}, largeErrors={:7.5f}, largeErrorsTot={:7.5f}, evals={:8.1f}, evalsTot={:8.1f}".format(
            method, 
            np.nanmean(methodStats["success"]), np.nanmean(methodStats["error"]),
            np.nanmean(methodStats["errorReduced"]), np.nanmean(methodStats["errorTot"]),
            np.nanmean(methodStats["largeErrors"]), np.nanmean(methodStats["largeErrorsTot"]),
            np.nanmean(methodStats["evals"]), np.nanmean(methodStats["evalsTot"])))
        

def _transform_parameters(params):
    k, p = params
    return np.exp(k), 1/(1+np.exp(-p))

def _LL_nbinom(params, data):
    k, p = _transform_parameters(params)
    return nbinom.logpmf(data, k, p).sum()
def _grad_LL_nbinom(params, data):
    return Gradient(_LL_nbinom)(params, data)
def _hess_LL_nbinom(params, data):
    return Hessian(_LL_nbinom)(params, data)

def test_find_CI():
    n = 100
    k, p = 5, 0.1
    data = np.random.negative_binomial(k, p, size=n)
    
    def LL(x):
        k, p = np.maximum(x, 1e-10)
        return nbinom.logpmf(data, k, p).sum()
    
    LLParallel = partial(_LL_nbinom, data=data)
    
    def nLLParallel(params):
        return -LLParallel(params)
    
    x0 = [1, 0.5]
    
    def nLL(params):
        return -LL(params)
    
    def function(params):
        return np.prod(params)
    
    functionJac = Gradient(function)
    functionHess = Hessian(function)
    
    jac = Gradient(LL)
    hess = Hessian(LL)
    
    result = op.differential_evolution(nLL, [(0, 10), (0, 1)])
    resultParallel = op.minimize(nLLParallel, x0)
    print("Estimate: k={:5.3f}, p={:5.3f}".format(*result.x))
    
    CI1 = find_CI_new(result.x, LL)
    print("All CIs:", CI1)
    CI2 = find_CI_new(resultParallel.x, LLParallel, parallel=True)
    print("All CIs parallel:", CI2)
    
    CIF0 = find_function_CI(result.x, lambda x: x[0], LL)
    print("CI for the first parameter (via function interface):", CIF0)
    
    CIF = find_function_CI(result.x, function, LL)
    print("CI for the product of all parameters:", CIF)
    CIF = find_function_CI(result.x, function, LL, logLJac=jac, logLHess=hess)
    print("CI for the product of all parameters:", CIF)
    CIF = find_function_CI(result.x, function, LL, functionJac=functionJac, functionHess=functionHess)
    print("CI for the product of all parameters:", CIF)
    CIF = find_function_CI(result.x, function, LL, logLJac=jac, logLHess=hess, functionJac=functionJac, functionHess=functionHess)
    print("CI for the product of all parameters:", CIF)
    
    CI3 = find_CI_new(result.x, LL, return_full_results=True)
    print("All CIs, full results:", CI3)
    CI4, success = find_CI_new(result.x, LL, return_success=True)
    print("All CIs and success:", CI4, success)
    CI5 = find_CI_new(result.x, LL, None, None, 1)
    print("Second parameter:", CI5)
    CI6 = find_CI_new(result.x, LL, None, None, None, 1)
    print("Upper bounds:", CI6)
    CI7 = find_CI_new(result.x, LL, None, None, None , [1, [False, True]])
    print("Upper bound for the first, both bounds for the second:", CI7)
    CI8 = find_CI_new(result.x, LL, None, None, [0, 1], [1, -1])
    print("Upper bound for the first, lower for the second:", CI8)
    CI9 = find_CI_new(result.x, LL, None, None, 0, 0)
    print("Lower bound for the first:", CI9)
    
    
if __name__ == '__main__':
    methods = ["Wald", "RVM", "RVM_psI", "bisection", "mixed_min", "constrained_max", "binsearch", "VM",  "gridsearch"]
    
    """
    # Uncomment this to test the H14 model, which requires additional R files.
    
    try:
        from rpy2 import robjects
        test_H14()
    except ImportError:
        raise ImportError("The package rpy2 must be installed in order to test "
                          "the H14 model.")
    """
    test_find_CI()
    benchmark(methods, 200, dataN=50, mode="11cx") 
    benchmark(methods, 200, dataN=10000, mode="11") 
    benchmark(methods, 200, dataN=10000, mode="3") 
    sys.exit()



'''
'''
import warnings
import traceback
import sys
from functools import partial
from multiprocessing import Pool
from itertools import count as itercount
import numdifftools as nd

import numpy as np
import scipy.optimize as op
from scipy.stats import chi2
from scipy.optimize._trustregion_exact import IterativeSubproblem

from vemomoto_core.tools.doc_utils import inherit_doc



def __round_str(n, full=6, after=3):
    if np.isfinite(n) and n != 0:
        digipre = int(np.log10(np.abs(n)))+1
        after = min(max(full-digipre-1, 0), after)
    return ("{:"+str(full)+"."+str(after)+"f}").format(n)

def is_negative_definite(M, tol=1e-5):
    """
    Returns whether the given Matrix is negative definite.
    
    Uses a Cholesky decomposition.
    """
    try:
        r = np.linalg.cholesky(-M)
        if np.isclose(np.linalg.det(r), 0, atol=tol):
            return False
        return True
    except np.linalg.LinAlgError:
        return False

class FlexibleSubproblem():
    """
    Class representing constained quadratic subproblems
    """
    def __init__(self, x, fun, jac, hess, hessp=None,
                 k_easy=0.1, k_hard=0.2):
        """
        ``k_easy`` and ``k_hard`` are parameters used
        to determine the stop criteria to the iterative
        subproblem solver. Take a look at the IterativeSubproblem class
        for more info.
        """
        self.x = x
        self.fun = fun
        self.jac = jac
        self.hess = hess
        self.hessp = hessp
        self.k_easy = k_easy
        self.k_hard = k_hard
        
    def solve(self, radius, positiveDefinite=None, tol=1e-5, jac0tol=0, *args, **kwargs):
        if radius == 0:
            return np.zeros_like(self.jac(self.x)), True
        
        subproblem = IterativeSubproblem(self.x, self.fun, self.jac, self.hess, 
                                         self.hessp, self.k_easy, self.k_hard)
        
        # IterativeSubproblem runs into issues, if jac is too small
        if subproblem.jac_mag <= max(subproblem.CLOSE_TO_ZERO, jac0tol):
            j = self.jac(self.x)
            h = self.hess(self.x)
            if positiveDefinite is None:
                positiveDefinite = is_negative_definite(-h, tol)
            if positiveDefinite:
                x = np.zeros(j.size)
            else:
                vals, vecs = np.linalg.eigh(h)
                x = vecs[:,np.argmin(np.real(vals))]
                x *= radius
            jNorm = np.linalg.norm(j)
            if jNorm:
                x_alt = j / jNorm * radius
                if (np.linalg.multi_dot((x, h, x)) + np.dot(j, x) 
                    > np.linalg.multi_dot((x_alt, h, x_alt)) + np.dot(j, x_alt)):
                    return x_alt, True
            return x, True
        return subproblem.solve(radius, *args, **kwargs)

class Flipper:
    """
    Flips an array in a specified component
    """
    def __init__(
            self, 
            index: "index in which the result should be flipped",
            ):
        self.index = index
    def __call__(self, x):
        if not hasattr(x, "__iter__"):
            return x
        index = self.index
        x = x.copy()
        x[index] *= -1
        if x.ndim > 1:
            x[:, index] *= -1
        return x

def get_independent_row_indices(M: "considered matrix", 
                                jac: "ordering vector" = None, 
                                tol: "numerical tolerance" = None
                                ) -> "boolean array of linearly independent rows":
    """
    Returns a boolean array arr with arr[i]==True if and only if
    the i-th row of M is linearly independent of all other rows j with 
    arr[j]==True.
    The vector jac provides an ordering of the returned indices. 
    If M[i] and M[j] are linearly dependent, then arr[i] will be True if
    jac[i] >= jac[j]. Otherwise, arr[i] will be False and arr[j] will be True.
    """
    n = M.shape[0]
    result = np.ones(n, dtype=bool)
    
    if jac is not None:
        indices = np.argsort(np.abs(jac))[::-1]
    else:
        indices = np.arange(n, dtype=int)
    
    rank = 0
    for i in range(n):
        if np.linalg.matrix_rank(M[indices[:i+1]], tol=tol) > rank:
            rank += 1
        else:
            result[indices[i]] = False
    
    if not result.any():
        result[indices[0]] = True
    return result

class CounterFun:
    """
    Counts how often a function has been called
    """
    def __init__(self, fun):
        self.fun = fun
        self.evals = 0
    def __call__(self, *args, **kwargs):
        self.evals += 1
        return self.fun(*args, **kwargs)


STATUS_MESSAGES = {0:"Success",
                   1:"Result on discontinuity",
                   2:"Result is unbounded",
                   3:"Iteration limit exceeded",
                   4:"Ill-conditioned matrix",
                   5:"Unspecified error"
                   }

def find_CI_bound(x0, fun, index, direction, 
        jac = None, 
        hess = None, 
        alpha = 0.95,
        fun0 = None, 
        jac0 = None, 
        hess0 = None, 
        customTarget = None,
        nmax = 200, 
        nchecks = 65,
        apprxtol = 0.5,
        resulttol = 1e-3,
        singtol = 1e-4, 
        minstep = 1e-5, 
        radiusFactor = 1.5,
        infstep = 1e10, 
        maxRadius = 1e4,
        disp = False, 
        track_x = False, 
        track_f = False):
    """Finds an end point of a profile likelihood confidence interval.
    
    Parameters
    ----------
    x0 : float[]
        Maximum likelihood estimate (MLE) of the paramters.
    fun : callable
        Log-likelihood function.
    index : int
        Index of the parameter of interest.
    direction : int or bool
        If ``<=0``, the lower end point of the confidence interval is sought, 
        else the upper end point is sought.
    jac : callable
        Gradient of :py:obj:`fun`. If `None`, it will be computed based on
        finite differences using numdifftools.
    hess : callable
        Hessian of :py:obj:`fun`. If `None`, it will be computed based on
        finite differences using numdifftools.
    alpha : float
        Desired confidence level. Must be in ``(0,1)``
    fun0 : float
        log-likelihood at the MLE.
    jac0 : float[]
        Gradient of the log-liekelihood at the MLE.
    hess0 : float[][]
        Hessian of the log-likelihood at the MLE.
    customTarget : float
        Custom target log-likelihood l*. If this is given, :py:obj:`alpha` will
        be ignored.
    nmax : int
        Maximal number of iterations.
    nchecks : int
        Maximal number of trust-region changes per iteration.
    apprxtol : float
        Relative tolerance between :py:obj:`fun` and its approximation.
    resulttol : float
        Tolerance of the result (``fun`` and ``norm(jac)``).
    singtol : float
        Tolerance for singularity checks.
    minstep : int
        Controls the minimal radius of the trust region. 
    radiusFactor : float
        Controls how quickly the trust region decreases. Must be in ``[1, 2]``.
    infstep : float
        Stepsize after which a parameter is deemed unestimbale.
    maxRadius : float
        Rradius of the trust region in the last iteration.
    disp : bool
        Whether to print a status message in each iteration.
    track_x : bool
        Whether to return the parameter trace.
    track_f : bool
        Whether to return the log-likelihood trace.
    
    """
    nuisance_minstep = 1e-3
    
    # ----- PREPARATION --------------------------------------------------------
    if jac is None:
        jac = nd.Gradient(fun)
    if hess is None:
        hess = nd.Hessian(fun)
    
    
    # Make sure arguments and returned values are of correct type
    index = int(index)  # in case the index is not given as integer
    x0 = np.asarray(x0) # in case x0 is not given as numpy array
    
    forward = direction > 0
    
    # wrap the functions so that the correct end point is found
    jac2 = jac
    hess2 = hess
    if not forward:
        fun2 = fun
        flip = Flipper(index)
        fun = lambda x: fun2(flip(x)) 
        jac = lambda x: flip(np.asarray(jac2(flip(x))))
        hess = lambda x: flip(np.asarray(hess2(flip(x))))
        x0 = x0.copy()
        x0[index] *= -1
    else: 
        # this is just to ensure the results are traced and returned correctly
        flip = lambda x: x
        jac = lambda x: np.asarray(jac2(x))
        hess = lambda x: np.asarray(hess2(x))
    
    # initial values
    if fun0 is None:
        fun0 = fun(x0)
    if jac0 is None:
        jac0 = jac(x0)
    else:
        jac0 = np.asarray(jac0)
    if hess0 is None:
        hess0 = hess(x0)
    else:
        hess0 = np.asarray(hess0)
    
    # flip initial values if necessary
    if not forward:
        jac0[index] *= -1
        hess0 = flip(hess0)
        
    # compute the target
    if customTarget is None:
        target = fun0 - chi2.ppf(alpha, df=1)/2
    else:
        target = customTarget
    
    i: "current iteration" = 0
    x_track: "trace of parameter values" = []
    f_track: "trace of log-likelihood values" = []
    maximizing: "flag controlling whether we are maximizing f w.r.t. x[index]" = False
    nuisanceRadius: "search radius for unconstrained subproblems" = 1
    
    # counters to keep track how often functions have been called
    fun = CounterFun(fun)
    jac = CounterFun(jac)
    hess = CounterFun(hess)
    
    multi_dot = np.linalg.multi_dot
    
    is_negative_definite_ = partial(is_negative_definite, tol=singtol)
    
    def fPredicted_fun():
        """
        Taylor approximation of f
        """
        dx = xTmp-x
        return f + np.dot(J, dx) + multi_dot((dx, H, dx))/2
    
    def JPredicted_fun():
        """
        Taylor approximation of jac
        """
        return J + np.dot(H, xTmp-x)
    
    norm: "applied norm" = lambda arr: np.sqrt(np.mean(arr*arr))
    
    def jac_approx_precise(JPredicted_, JActual_, J_, bound=apprxtol):
        """
        Returns True, iff the predicted gradient is close to the
        actual one or, even better, closer to the target than predicted
        """
        # note how far we are from the target likelihood
        if (not searchmode=="normal" or maximizing) and not closeToDisc:
            return True
            
        return norm(JActual_) / norm(J_) <= bound or norm(JPredicted_-JActual_) / norm(J_) <= bound
    
    # TODO Accept improvements only then always, if they are better by a certain constant (compare values - line search)
    def f_approx_precise(fPredicted, fActual, bound=apprxtol):
        """
        Returns True, iff the predicted log-likelihood value is close to the
        actual one or, even better, closer to the target than predicted
        """
            
        # steps to negative predicted values cannot occur in theory and are 
        # either due to numerical issues or because we tried a very long step. 
        # Therefore, we reject these steps right away, if fActual is below the 
        # target.
        if fPredicted < target - resulttol and f > target and fActual < target:
            return False
        
        # return True if result is better than expected
        if fActual > fPredicted and xiTmp >= 0:
            return True
        
        # if we are maximizing w.r.t. the nuisance parameters, 
        if searchmode=="max_nuisance" or searchmode=="max_nuisance_const" or maximizing:
            # we do not accept steps that are worse than the current stage
            # (we guarantee elsewhere that it IS possible to increase f)
            if fActual < f: 
                if not relax_maximum_constraint or np.abs(fPredicted-fActual) > resulttol:
                    return False
            bound *= 2
            if relax_maximum_constraint:
                bound *= 2
        elif searchmode=="normal":
            # if we are outside the admissible region, we enforce that
            # we get closer to it again
            if fActual < target and np.abs(fActual-target) > np.abs(target-f):
                return False
            elif np.abs(fActual-target) < resulttol and np.abs(fPredicted-target) < resulttol:
                return True
            elif fActual > f > target:
                # we relax the precision requirement, if the step is not of
                # disadvantage and brings us forward (we value progression in 
                # the paramater of interest higher than a maximal likelihood)
                bound *= 2
                
        
        # we require that the error relative to the distance from the target
        # is not larger than the given bound.
        return np.abs(fPredicted-fActual) / np.abs(target-f) <= bound
    
    def test_precision() -> "True, if appxomiations are sufficiently precise":
        """
        Tests the precision of the Taylor approximation at the suggested 
        parameter vector.
        """
        fActual = fun(xTmp)
        fPredicted = fPredicted_fun()
        
        if not f_approx_precise(fPredicted, fActual):
            return False, fActual, None
        
        if np.abs(f-target) > resulttol and not closeToDisc: 
            return True, fActual, None
        
        JActual = jac(xTmp)
        JPredicted = JPredicted_fun() 
        
        if free_params_exist: 
            JActual_ = JActual[~free_considered]
            JPredicted_ = JPredicted[~free_considered]
            
            # the jacobian shall be mostly precise w.r.t. the changed parameters.
            # However, it shoudl not be totally imprecise with respect to the
            # remaining parameters, either.
            jac_precise = (jac_approx_precise(JPredicted_, JActual_, J_) and
                           jac_approx_precise(JPredicted[considered], 
                                              JActual[considered],
                                              J[considered], 
                                              20*apprxtol))
        else:
            JActual_ = JActual[considered]
            JPredicted_ = JPredicted[considered]
            jac_precise = jac_approx_precise(JPredicted_, JActual_, J_)
        
        return jac_precise, fActual, JActual
        
        
    
    def is_concave_down():
        """
        Checks whether the profile likelihood function is concave down.
        """
        return a < 0
    
    if disp:
        def dispprint(*args, **kwargs):
            print(*args, **kwargs)
    else:        
        def dispprint(*args, **kwargs):
            pass
    
    
    # variable definitions -------------------------------------------------
    x:      "parameter vector at current iteration"      = x0
    xTmp:   "potential new parameter vector"
    xi:     "parameter of interest at current iteration" = x0[index]
    f:      "log-likelihood at current iteration"        = fun0
    J:      "gradient at current iteration"              = jac0
    H:      "Hessian at current iteration"               = hess0
    maxFun: "maximal log-likelihood"                     = fun0
    xi_max: "maximal value for the parameter of interest with f(xi, x_) >= f*"    = xi
    x_max:  "parameter vector where xi_max leads to a log-likelihood value >= f*" = x
    considered:     "Array indexing nuisance parameters" = np.ones_like(x0, dtype=bool)
    considered[index]  = False
    consideredOriginal = considered.copy()
    hessConsidered: "Array indexing nuisance Hessian"    = np.ix_(considered, considered)
    H_:     "nuisance Hessian"                           = H[hessConsidered]
    H0_:    "nuisance entries of Hessian row corresponding to the parameter of interest" = H[index][considered]
    discontinuity_counter: "steps until discontinuous parameters are released" = 0
    maximizing: "flag to memorize whether the target has been changed temporarily" = False
    resultmode: "denotes whether or what kind of result has been found" = "searching"
    radius: "radius of the trust region in the last iteration" = np.inf
    closeToDisc: "True if we ar close to a discontinuity" = False
    xi_hist = []
    f_hist = []
    
    
    subproblem = FlexibleSubproblem(None, 
             lambda x: -f, 
             lambda x: -J_ - (xiTmp-xi)*H0_,
             lambda x: -H_
             )
    
    
    a = 0                        # just to prevent a possible name error...
    xiRadius = nuisanceRadius
        
    # Wrapper to catch potential errors without braking down completely
    try:    
        # ----- ITERATIONS -----------------------------------------------------
        for i in range(1, nmax+1):
            
            redoIteration = False
            ballsearch = False
            jacDiscont = False
            free_params_exist = False
            free_considered = ~considered
            infstepTmp = max(infstep, 10*np.abs(x0[index]))
            #closeToDisc = False
            
            if track_x:
                x_track.append(flip(x))
            if track_f:
                f_track.append(f)
            
            # if discontinuities are found, some parameters are held constant
            # for a while. The discontinuity_counter keeps track of the time
            if discontinuity_counter:
                discontinuity_counter -= 1
            else:
                considered = consideredOriginal.copy()
                hessConsidered = np.ix_(considered, considered)
                relax_maximum_constraint = False
            
            k = -1
            
            # determine approximation
            J_ = J[considered]
            xi = x[index]
            H0 = H[index]
            H0_ = H0[considered]
            H00 = H[index, index]
            H_ = H[hessConsidered]
            
            # Determine search mode:
            # max_nuisance:  set xi to specific value, maximize other parameters
            # max_nuisance_const:  hold xi constant, maximize other parameters
            # normal:        search x with f(x) = f*, J(x) = 0
            # binary search: get back to the adissible region with a binary 
            #                search
            counter = 0
            while True:
                counter += 1
                if counter > 10:
                    raise RuntimeError("Infinite loop", f, J, H, considered, 
                                       free_considered)
                
                if is_negative_definite_(H_):
                    H_inv = np.linalg.inv(H_)
                    searchmode = "normal"
                    break
                
                free_considered_ = ~get_independent_row_indices(H_, J_, 
                                                                tol=singtol)
                free_considered = ~considered
                free_considered[considered] = free_considered_
                free_params_exist = free_considered_.any()
                
                if not free_params_exist:
                    searchmode = "max_nuisance"
                    break
                
                H_dd = H_[np.ix_(~free_considered_, ~free_considered_)]
                
                # special case for H_ = 0
                if H_dd.size == 1 and np.allclose(H_dd, 0, atol=singtol):
                    # though H_ is singular, we set its inverse to a very large
                    # value to prevent NaNs that would mess up the result
                    H_ddInv = np.array([[max(-infstep/singtol, 1/H_dd[0,0])]])
                elif not is_negative_definite_(H_dd):
                    free_considered = ~considered
                    free_considered_[:] = False
                    searchmode = "max_nuisance"
                    break
                else:
                    H_ddInv = np.linalg.inv(H_dd)
                
                
                
                H0_d = H0_[~free_considered_]
                J_d = J_[~free_considered_]
                
                H_inv = H_ddInv
                H0_ = H0_d
                J_ = J_d
                H_ = H_dd
                
                searchmode = "normal"
                break
                
            # ----- NORMAL SEARCH ----------------------------------------------   
            while searchmode == "normal":
                
                a = (H00 - multi_dot((H0_, H_inv, H0_)))/2
                p = J[index] - multi_dot((H0_, H_inv, J_))
                fPL0: "current value on profile log-likelihood" 
                fPL0 = f - multi_dot((J_, H_inv, J_))/2
                
                if maximizing:
                    if (is_concave_down() 
                            and not (np.allclose(a, 0, atol=singtol) and 
                                     np.allclose(((fPL0-target)/p)*(a/p), 0, 
                                                 atol=singtol))
                            ) or f < target:
                        maximizing = False
                    else:
                        targetTmp = max(fPL0 + 1, (maxFun+f)/2)
                
                if not maximizing:
                    targetTmp = target
                    
                q = fPL0 - targetTmp
                
                if a == 0:
                    sqroot = np.nan
                else:
                    sqroot = (p/a)**2 - 4*q/a 
                
                
                # if desired point is not on approximation
                if sqroot < 0:
                    if (not np.allclose(a, 0, atol=singtol) 
                            and get_independent_row_indices(H, J, tol=singtol
                                                            )[index]):
                        if not maximizing and a<0 and f < targetTmp:
                            warnings.warn("The current function value is not on the approximation. "
                                          + "This may be due to a too large singtol. "
                                          + "The algorithm may not converge. Index=" 
                                          + str(index) + "; forward=" 
                                          + str(forward) + "; a=" 
                                          + str(a) + "; f=" + str(f)
                                          + "; root=" + str(sqroot))
                        
                        assert not maximizing
                        
                        if not is_concave_down() and f >= target:
                            # if beyond or very close to a local min
                            if p >= 0 or -p/a < singtol:
                                maximizing = True
                                continue
                            else:
                                # jump over minimum by factor 2
                                p *= 2
                            
                        sqroot = 0
                    # if f is approximately constant, we do not trust the 
                    # quadratic approximation too much and jsut try a very long
                    # step, even if it goes beyond the predicted admissible 
                    # region
                    else:
                        a = 0
                        sqroot = np.inf
                else:    
                    sqroot = np.sqrt(sqroot)
                
                # if Hessian is singular in the considered component
                # this control is important to prevent numerical errors if
                # |a| << 1 and the summands of the root are equal order of 
                # magnitude
                if (np.allclose(a, 0, atol=singtol) and 
                        np.allclose((q/p)*(a/p), 0, atol=singtol)):
                    if (maximizing and q*p>0) or (not maximizing and 
                                                    f >= target and q*p > 0):
                        maximizing = not maximizing
                        continue
                    else:
                        xiTmp = xi - q/p
                # control for NaN case
                elif p == 0 and a == 0: 
                    xiTmp = np.inf
                else:
                    xi1 = xi + (-p/a + sqroot) / 2
                    xi2 = xi + (-p/a - sqroot) / 2
                    
                    # check which root is closer to current value
                    if is_concave_down(): #not maximizing
                        assert not maximizing
                        xiTmp = max(xi1, xi2)
                    elif p*a > 0 and not maximizing and f > target:
                        maximizing = True
                        continue
                    elif maximizing and p/a < -singtol:
                        maximizing = False
                        continue
                    elif maximizing:
                        xiTmp = max(xi1, xi2)
                    else:
                        if np.abs(xi-xi1) < np.abs(xi-xi2): 
                            xiTmp = xi1
                        else:
                            xiTmp = xi2
                
                # if likelihood is independent of parameter
                if np.abs(xiTmp-xi) >= infstep * (1-singtol):
                    
                    
                    # try a very large xi value
                    xiTmp = xi + infstepTmp
                    xTmp = x.copy()
                    xTmp[index] = xiTmp
                    xTmp[~free_considered] += -np.dot(H_inv, H0_*(xiTmp-xi)+J_)
                    fActual = fun(xTmp)
                    
                    # if still larger than target
                    if fActual >= target:
                        JActual = jac(xTmp)
                        resultmode = "unbounded"
                        break 
                    
                    if not f > target:
                        if norm(J_) > singtol:
                            searchmode = "max_nuisance"
                            break 
                        
                        searchmode = "binary_search"
                        break 
                    xiTmp = xiTmp/4 + 3*xi/4
                
                xTmp = x.copy()
                xTmp[index] = xiTmp
                xTmp[~free_considered] += -np.dot(H_inv, H0_*(xiTmp-xi)+J_)
                
                xx = x[~free_considered]
                upperRadius = np.linalg.norm(xx-xTmp[~free_considered])
                
                # Test precision
                stop, fActual, JActual = test_precision()
                if stop:
                    radius = max(radius, upperRadius)
                    break
                
                if radius >= upperRadius:
                    lowerRadius = upperRadius/2
                else:
                    lowerRadius = radius
                
                if upperRadius > 0:
                    radius = upperRadius
                
                tmpRadius = lowerRadius
                xiStep = xiTmp-xi
                
                if tmpRadius == 0:
                    xiTmp = xi + min(xiStep/2, infstep)
                else:
                    xiTmp = xi + xiStep*np.power(tmpRadius/radius, np.log(2)/np.log(radiusFactor))
                
                ballsearch = True
                break
                    
            if not resultmode=="searching":
                x = xTmp
                J = JActual
                JActual_ = JActual[considered]
                f = fActual
                break
            
            if searchmode == "normal" and free_params_exist:
                J_new = norm(JPredicted_fun()[considered])
                J_current = norm(J[considered])
                
                # consider adjusted scenario with xiTmp = 0
                J_f = J[considered][free_considered_]
                H_df = H[hessConsidered][np.ix_(~free_considered_, free_considered_)]
                J_new2 = J_f - multi_dot((H_df.T, H_ddInv, J_d))
                J_new2 = np.linalg.norm(J_new2)/np.sqrt(J_.size)
                
                # if ignoring the free parameters hinders us from making an 
                # improvement w.r.t. the gradient
                abstol = max(np.log2(np.abs(xiTmp-xi))*np.abs(f-target)
                             /np.abs(maxFun-target), resulttol)
                if (J_new > abstol and J_new/J_current > apprxtol) or (
                        J_new2 > abstol and J_new2/J_current > apprxtol):
                    free_considered = ~considered
                    free_considered_[:] = False
                    free_params_exist = False
                    J_ = J[considered]
                    H_ = H[hessConsidered]
                    H0_ = H0[considered]
                    searchmode = "max_nuisance"
            
            
            counter = 0
            repeat = True
            while repeat:
                repeat = False
                
                if counter > 10:
                    raise RuntimeError("Infinite loop", f, J, H, considered, 
                                       free_considered)
                
                # ----- CONSTRAINED MAXIMIZATION -----------------------------------
                if searchmode == "max_nuisance" or searchmode == "max_nuisance_const":
                    xiTmp = xi
                    tmpRadius = nuisanceRadius * np.sqrt(J_.size)
                    ballsearch = True
                    
                    if not searchmode == "max_nuisance_const":
                        if f > target:
                            xiTmp += xiRadius
                        else:
                            searchmode = "max_nuisance_const"
                
                # ----- BALL SEARCH ------------------------------------------------   
                if ballsearch:
                    increaseTrustRegion = True
                    xiPrev = xi
                    nAdjustement = 5
                    for k in range(nchecks):
                        
                        # it can happen that the radius is so conservative that
                        # the nuisance parameters cannot be maximized 
                        # sufficiently far to keep f in the admissible region
                        nRatioAdjustement = 10
                        for l in range(nRatioAdjustement):
                            #print("tmpRadius", tmpRadius)
                            xTmp_diff, _ = subproblem.solve(tmpRadius, 
                                                            searchmode=="normal",
                                                            jac0tol=singtol)
                            if xi > xiTmp and f>target:
                                # Somethong went wrong. Print debug information.
                                warnings.warn("Something went wrong. Please report this error "
                                              "and the information below to the bugtracker. "
                                              "Presumably the result is correct anyway.")
                                try:
                                    print("searchmode, maximizing", searchmode, maximizing)
                                    print("a, p =", a, ",", p)
                                    print("q, fPL0 =", q, ",", fPL0)
                                    print("f, target, targetTmp =", f, ",", target,",",  targetTmp)
                                    print("xiTmp, xi =", xiTmp, ",", xi)
                                    print("xi1, xi2 =", xi1, ",", xi2)
                                    print("xiStep, infstep =", xiStep, ",", infstep)
                                    print("tmpRadius, radius =", tmpRadius, ",", radius)
                                    print("index, direction, k, l =", index, forward, k, l )
                                    print("xTmp[index]", xTmp[index])
                                    print("---------------------")
                                except Exception:
                                    pass
                            xTmp = x.copy()
                            xTmp[index] = xiTmp
                            xTmp[~free_considered] += xTmp_diff
                            fPredicted = fPredicted_fun()
                            # if we maximize w.r.t. the nuisance parameters,
                            # we want to assure that f gets increased to get
                            # back to the likelihood ridge
                            if (not searchmode == "max_nuisance" and 
                                    not searchmode == "max_nuisance_const"):
                                break
                            precise, fActual, JActual = test_precision()
                            if fPredicted >= f:
                                break
                            # if the step is super close and f still does not 
                            # increase, there must be a numerical issue. We
                            # leave the treatment of this issue to the procedures
                            # below.
                            elif (np.isclose(xiTmp, xi, atol=minstep/100) and tmpRadius<minstep/100):
                                break
                            
                            if l == nRatioAdjustement-2:
                                xiPrev = xiTmp
                                xiTmp = xi
                            else:
                                # we either increase the search radius for the
                                # nuisance parameters or decrease the search
                                # radius for the parameter of interest dependent
                                # on whether the approximation is precise and
                                # we can increase the trust region or not
                                if increaseTrustRegion and precise and tmpRadius < maxRadius*np.sqrt(xTmp_diff.size):
                                    tmpRadius *= 2
                                else:
                                    # if the error occurs even though xi is 
                                    # constant, there must be an error in the 
                                    # approximate solution and we decrease the
                                    # search radius
                                    if np.isclose(xiTmp, xi, atol=minstep/100):
                                        tmpRadius /= 5
                                    xiTmp = (xi+xiTmp) / 2
                                    
                        # if maximizing w.r.t. the nuisance parameters,
                        # adjust the trust region
                        if (searchmode == "max_nuisance" or 
                                searchmode == "max_nuisance_const"):
                            if precise:
                                # update trust region
                                nuisanceRadius = tmpRadius / np.sqrt(J_.size)
                                # update xiRadius only if we have changed xi
                                if searchmode == "max_nuisance":
                                    if xi != xiTmp:
                                        xiRadius = max(xiTmp-xi, minstep)
                                    else:
                                        # if we had to set xiRadius to 0, we set 
                                        # it to the last non-zero value
                                        xiRadius = max(xiPrev-xi, minstep)
                                    # since the xiRadius cannot recover on its
                                    # own, we always gradually increase it
                                    if 2*xiRadius < nuisanceRadius:
                                        xiRadius *= 2
                                
                                # if we have not decreased the trust region 
                                # before, try to increase it
                                if increaseTrustRegion and k < nAdjustement:
                                    xTmp2 = xTmp
                                    fActual2 = fActual
                                    xiTmp = xi + 2*xiRadius
                                    tmpRadius *= 2
                                    tmpRadius = min(maxRadius*np.sqrt(xTmp_diff.size),
                                                    tmpRadius)
                                    continue
                                else:
                                    stop = True
                            else:
                                stop = False
                                if increaseTrustRegion: 
                                    if k == 0:  
                                        # if in the first iteration, note that 
                                        # we are decreasing the trust region
                                        increaseTrustRegion = False
                                    else:
                                        # set xTmp to the last precise value
                                        fActual = fActual2
                                        xTmp = xTmp2
                                        xiTmp = xTmp[index]
                                        stop = True
                        else:    
                            stop, fActual, JActual = test_precision()
                        
                        if searchmode == "normal":
                            
                            if stop:
                                # if the approximation is precise, we may increase
                                # the radius, but save the largest current result
                                # to fall back to
                                xTmp2 = xTmp
                                fActual2 = fActual
                                lowerRadius = tmpRadius
                            else:
                                upperRadius = tmpRadius
                                
                            # if we have to decrease the trust region
                            if upperRadius <= lowerRadius and not stop:
                                # continue with the continuity checks and
                                # cut the trust trust region by half
                                pass
                            # if precision of geometric binary search is as 
                            # desired (first equality necessary to prevent NaNs)
                            elif upperRadius==0 or upperRadius/lowerRadius <= 2:
                                # stop iteration and set xTmp to the optimal 
                                # value
                                if lowerRadius > 0:
                                    radius = lowerRadius
                                fActual = fActual2
                                xTmp = xTmp2
                                xiTmp = xTmp[index]
                                break
                            else:
                                # update xiTmp and radius
                                tmpRadius = np.sqrt(lowerRadius*upperRadius)
                                xiTmp = xi + xiStep*np.power(tmpRadius/radius, np.log(2)/np.log(radiusFactor))
                                continue
                        elif stop:
                            break
                        
                        minstepTmp = max(minstep / max(norm(J_), 1), 1e-12)
                        
                        # if the step is too small
                        diffs = xTmp - x
                        if not stop and np.allclose(xTmp, x, atol=minstepTmp, rtol=1e-14):
                            
                            # if we are not really on a discontinuity but only 
                            # on a local maximum with a non-negative definite
                            # Hessian, we may just release the constraint
                            # that we need to get to a larger function value
                            if (not searchmode=="normal" or maximizing) and not closeToDisc: 
                                relax_maximum_constraint = True
                                discontinuity_counter = 20
                                if test_precision()[0]:
                                    break
                                
                            # check parameter of interest first
                            xTmp = x.copy()
                            xTmp[index] += diffs[index]
                            
                            # if small change in parameter of interest makes 
                            # approximation imprecise
                            if diffs[index] and not test_precision()[0]:
                                
                                # make sure discontinuity happens on 
                                # profile function
                                if norm(J[considered]) > resulttol:
                                    relax_maximum_constraint = False
                                    searchmode = "max_nuisance_const"
                                    closeToDisc = True
                                    xiDisc = xi
                                    repeat = True
                                    radius = tmpRadius
                                    break
                                
                                fActual = fun(xTmp)
                                fCloseTarget = (np.allclose(target, f, 
                                                            atol=resulttol) and
                                                not np.allclose(target, fActual, 
                                                                atol=resulttol))
                                
                                # if step is favourable or benign, we accept 
                                # depending on the distance measure this case
                                # will never occur, as favourable steps may 
                                # always be classified as precise
                                if fActual >= target or f < fActual:
                                    pass
                                # if the target is on the jump
                                elif ((f > target or fCloseTarget)
                                      and fActual < target):
                                    # stop here
                                    fPredicted = fPredicted_fun()
                                    resultmode = "discontinuous"
                                    xTmp = x
                                    JActual = J
                                else:
                                    # get back to the feasible region by 
                                    # conducting a binary search
                                    searchmode = "binary_search"
                                    
                                dispprint("iter {:3d}{}: ***discontinuity of {} in x_{} at {}***".format(
                                            i, ">" if forward else "<", f-fActual, 
                                            index, x))
                                break
                            
                            # check for discontinuities in nuisance parameters
                            discontinuities = np.zeros_like(free_considered)
                            discontinuities_rejected = discontinuities.copy()
                            if not f_approx_precise(fPredicted, fActual):
                                if diffs[index]:
                                    fPredicted = fPredicted_fun()
                                    fActual = fun(xTmp)
                                else:
                                    fPredicted = fActual = f
                                for l in np.nonzero(~free_considered)[0]:
                                    xTmp[l] += diffs[l]
                                    fActual_previous = fActual
                                    fActual = fun(xTmp)
                                    fPredicted = fPredicted_fun()
                                    if not f_approx_precise(fPredicted, fActual):
                                        discontinuities[l] = True
                                        if fActual < fActual_previous:
                                            discontinuities_rejected[l] = True
                                            xTmp[l] = x[l]
                                            fActual = fActual_previous
                                jacDiscont = False
                            # if the gradient is responsible for the rejection
                            # we require a smaller steap
                            elif np.abs(f-target) < resulttol and np.allclose(xTmp, x, rtol=minstepTmp/10):
                                for l in np.nonzero(~free_considered)[0]:
                                    xTmp[l] += diffs[l]
                                    JActual_previous = JActual
                                    JActual = jac(xTmp) 
                                    JPredicted = JPredicted_fun()
                                    jac_precise = jac_approx_precise(JPredicted[~free_considered], JActual[~free_considered], J[~free_considered]) 
                                    if not jac_precise:
                                        discontinuities[l] = True
                                        if not (JActual_previous[l] * J[l] >= 0 
                                                or diffs[l]*J[l] <= 0):
                                            discontinuities_rejected[l] = True
                                            xTmp[l] = x[l]
                                            JActual = JActual_previous
                                jacDiscont = True
                            

                            if discontinuities.any():
                                
                                varStr = "variables" if not jacDiscont else "gradient entries"
                                    
                                dispprint("iter {:3d}{}: ***index={}: discontinuities in {} {} at {}***".format(
                                            i, ">" if forward else "<", index, varStr,
                                            tuple(np.nonzero(discontinuities)[0]), x))
                                
                                redoIteration = discontinuities.sum() == discontinuities_rejected.sum()
                                
                                if discontinuities_rejected.any():
                                    discontinuity_counter = 25
                                    considered &= ~discontinuities_rejected
                                    free_considered |= discontinuities_rejected
                                    free_considered_ = free_considered[considered]
                                    hessConsidered = np.ix_(considered, considered)
                                    relax_maximum_constraint = False
                                
                                break
                        if tmpRadius > (radiusFactor**nchecks-k)*minstep:
                            tmpRadius /= 5
                            xiTmp = (4*xi+xiTmp) / 5
                        else:
                            tmpRadius /= radiusFactor
                            xiTmp = (xi+xiTmp) / 2
                    else:
                        xTmp = x
                        JActual = J
                        fActual = f
            
            if redoIteration:
                continue
                    
            # ----- BINARY SEARCH ----------------------------------------------
            if searchmode == "binary_search":
                tol = 0.01
                bin_min = x
                bin_max = x_max
                for k in range(nchecks):
                    xTmp = (bin_max+bin_min) / 2
                    fActual = fun(xTmp)
                    
                    if 0 <= fActual - target < tol:
                        break
                    
                    if fActual < target:
                        bin_min = xTmp
                    else:
                        bin_max = xTmp
                else: 
                    xTmp = bin_max
                JActual = jac(xTmp)
                        
                        
            # ----- UPDATES & CEHCKS -------------------------------------------
            if (xTmp == x).all() and not jacDiscont and resultmode=="searching":
                dispprint("iter {:3d}{}: ***no improvement when optimizing x_{} at {}***".format(
                            i, ">" if forward else "<", index, flip(x)))
                resultmode = "discontinuous"
            
            
            
            xiChange = xTmp[index] - xi
            xChange = norm(x-xTmp)
            
            # when we are maximizing the nuisance parameters, we may step 
            # farther than the allowedbound
            if xiChange >= infstep and f >= target:
                resultmode = "unbounded"
            elif (searchmode == "max_nuisance" or searchmode == "max_nuisance_const"
                  or maximizing) and (xiChange < nuisance_minstep 
                                      and f > target+resulttol):
                if (len(xi_hist) > 5 and (np.abs(np.array(xi_hist[-5:]) - xi) < nuisance_minstep).all() and
                    (np.abs(np.array(f_hist[-5:]) - f) < nuisance_minstep).all()):
                    
                    xTmp[index] += nuisance_minstep
                    fActual = fun(xTmp)
                    m = 0
                    fActual_previous = fActual
                    xTmp_previous = xTmp.copy()
                    while fActual > target and m < 1000:
                        fActual_previous = fActual
                        xTmp_previous = xTmp.copy()
                        xTmp[index] += nuisance_minstep * 2**m
                        fActual = fun(xTmp)
                        m += 1
                    
                    xTmp = xTmp_previous
                    fActual = fActual_previous
                    
                    m = -1
                    while fActual < target-50 and m > -10:
                        fActual_previous = fActual
                        xTmp_previous = xTmp.copy()
                        xTmp[index] += nuisance_minstep * 2**m
                        fActual = fun(xTmp)
                        m -= 1
                    
                    
                    JActual = jac(xTmp)
                    
                    dispprint("iter {:3d}{}: ***taking additional step in x_{} to avoid convergence issues. f={:6.3}***".format(
                        i, ">" if forward else "<", index, fActual))
                    for m in range(1, 5):
                        xi_hist[-m] += xTmp[index]-xi
                xi_hist.append(xTmp[index])
                f_hist.append(fActual)
            
            x = xTmp
            xi = xTmp[index]
            
            if JActual is None:
                JActual = jac(xTmp)
            JImprovement = np.log2(norm(J[considered])/norm(JActual[considered]))
            fImprovement = np.log2(np.abs((f-target)/(fActual-target)))
            
            J = JActual
            f = fActual
            JActual_ = JActual[considered]
            
            # reset maximal value
            if f > target and xi > xi_max:
                xi_max = xi
                x_max = x
            
            fPredicted = fPredicted_fun()
            
            if disp:
                radius_str = ""
                if searchmode == "normal":
                    if maximizing:
                        mode = "maximizing"
                    else:
                        mode = "searching"
                elif (searchmode == "max_nuisance" 
                      or searchmode == "max_nuisance_const"):
                    mode = "maximizing nuisance parameters"
                    radius_str = "; step={}; radius={}".format(
                                        __round_str(xiRadius), __round_str(nuisanceRadius))
                elif searchmode == "binary_search":
                    mode = "binary search"
                else:
                    mode = "other"
                
                ndiscont = np.sum(~considered)-1
                if ndiscont:
                    discontStr = "; ndisct={}".format(ndiscont)
                else: 
                    discontStr = ""
                    
                if free_params_exist and free_considered_.any():
                    mode += " with " + str(np.sum(free_considered_)
                                           ) + " free params"
                    jac_cnsStr = "; jac_cns={}".format(__round_str(
                            norm(JActual[~free_considered])))
                else:
                    jac_cnsStr = ""
                    
                dispprint(("iter {:3d}{}: x_{}_d={}; f_d={}; jac_d={}; " + 
                       "nsteps={:2d}; x_d={}; f_impr={}; jac_impr={}; " +
                       "f_e={}{}{}{} - {}").format(i, 
                                                 ">" if forward else "<", index, 
                                                 __round_str(xi-x0[index]), 
                                                 __round_str(f-target), 
                                                 __round_str(norm(JActual_)), k+2, 
                                                 __round_str(xChange), 
                                                 __round_str(fImprovement), 
                                                 __round_str(JImprovement), 
                                                 __round_str(fActual-fPredicted), 
                                                 jac_cnsStr, discontStr, 
                                                 radius_str, mode))
            
            
            if resultmode == "searching":
                resViol = max(norm(JActual_), np.abs(f-target))
                if resViol <= resulttol:
                    resultmode = "success"
                    break
            else:
                break
            
            if f > maxFun:
                dispprint("-> iter {:3d}{}: !!!found NEW MAXIMUM for x_{} of {:6.3f} (+{:6.3f}) at {}!!!".format(
                        i, ">" if forward else "<", index, f, f-fun0, flip(x)))
                maxFun = f
            
            if closeToDisc:
                if np.abs(xi - xiDisc) > minstep:
                    closeToDisc = False
            
            H = hess(x)
        
        else:
            resultmode = "exceeded"
    except np.linalg.LinAlgError:
        resultmode = "linalg_error"
        exc_type, exc_value, exc_traceback = sys.exc_info()
        traceback.print_exception(exc_type, exc_value, exc_traceback,
                                  file=sys.stdout)
    except Exception:
        resultmode = "error"
        exc_type, exc_value, exc_traceback = sys.exc_info()
        traceback.print_exception(exc_type, exc_value, exc_traceback,
                                  file=sys.stdout)

    
    if resultmode == "success":
        status = 0
        success = True
    elif resultmode == "discontinuous":
        status = 1
        success = False
    elif resultmode == "unbounded":
        status = 2
        success = True
    elif resultmode == "exceeded":
        success = False
        status = 3
    elif resultmode == "error" or resultmode == "linalg_error":
        try:        
            JActual_ = J_
        except UnboundLocalError:
            JActual_ = jac(x)[considered]
        success=False
        if resultmode == "error":
            status = 5
        else:
            status = 4
    
    dispprint(op.OptimizeResult(x=flip(x), 
                         fun=f,
                         jac=JActual_,
                         success=success, status=status,
                         nfev=fun.evals, njev=jac.evals, nhev=hess.evals, 
                         nit=i,
                         message=STATUS_MESSAGES[status]
                         ))
    
    if track_f:
        f_track.append(f)
    if track_x:
        x_track.append(flip(x))
    
    return op.OptimizeResult(x=flip(x), 
                             fun=f,
                             jac=JActual_,
                             success=success, status=status,
                             nfev=fun.evals, njev=jac.evals, nhev=hess.evals, 
                             nit=i,
                             x_track=np.array(x_track),
                             f_track=np.array(f_track),
                             message=STATUS_MESSAGES[status]
                             )

def __parallel_helper_fun(args):
    return args[0], args[3], find_CI_bound(*args[1:-1], **args[-1])

@inherit_doc(find_CI_bound)
def find_CI(x0, fun, jac=None, hess=None, indices=None, directions=None, alpha=0.95, 
            parallel=False, return_full_results=False, return_success=False,
            **kwargs):
    """Returns the profile likelihood confidence interval(s) for one or 
    multiple parameters.
    
    Parameters
    ----------
    indices : int[]
        Indices of the parameters of interest. If not given, all paramters
        will be considered.
    directions : float[][]
        Search directions. If not given, both end points of the confidence
        intervals will be determined. If given as a scalar, only lower
        end points will be returned if ``directions<=0`` and upper end points 
        otherwise. If given as an array, the confidence interval end points 
        specified in row ``i`` will be returned for parameter ``i``. Entries ``<=0``
        indicate that lower end points are desired, whereas positive entries 
        will result in upper end points.
    parallel : bool
        If ``True``, results will be computed in parallel using ``multiprocessing.Pool``.
        Note that this requires that all arguments are pickable.
    return_full_results : bool
        If ``True``, an ``OptimizeResult`` object will be returned for each 
        confidence interval bound. Otherwise, only the confidence interval 
        bounds for the parameters in question will be returned.
    return_success : bool
        If ``True``, an array of the same shape as the result will be returned
        in addition, indicating for each confidence interval bound whether it 
        was determined successfully. 
    **kwargs : keyword arguments
        Other keyword arguments passed to :py:meth:`find_CI_bound`. Look at 
        the documentation there.
        
    """
    x0 = np.asarray(x0)
    
    isScalarIndex = indices is not None and np.isscalar(indices)
    if indices is None:
        indices = np.arange(x0.size)
    elif isScalarIndex:
        indices = (int(indices),)
    else:
        indices = np.array(indices, dtype=int)
    
    isScalarDirection = directions is not None and np.isscalar(directions)
    if directions is None:
        directions = np.zeros((len(indices), 2))
        directions[:,0] = -1
        directions[:,1] = 1
    elif isScalarDirection:
        directions = np.full((len(indices), 1), directions)
    elif type(directions) == np.ndarray:
        if directions.ndim == 1:
            directions = directions[:,None]
    else:
        directions = [[i] if not hasattr(i, "__iter__") else i 
                      for i in directions]
    
    if "fun0" not in kwargs:
        kwargs["fun0"] = fun(x0)
    if "jac0" not in kwargs:
        if jac is None:
            kwargs["jac0"] = nd.Gradient(fun)(x0)
        else:
            kwargs["jac0"] = jac(x0)
    if "hess0" not in kwargs:
        if hess is None:
            kwargs["hess0"] = nd.Hessian(fun)(x0)
        else:
            kwargs["hess0"] = hess(x0)
    
    task_chain = []
    
    for i, index, directions_ in zip(itercount(), indices, directions):
        for direction in directions_:
            task_chain.append((i, x0, fun, index, direction, jac, hess, 
                               alpha, kwargs))
    
    if parallel:
        with Pool() as pool:
            result_list = list(pool.map(__parallel_helper_fun, task_chain))
    else:
        result_list = list(map(__parallel_helper_fun, task_chain))
    
    result = [[] for _ in range(min(len(indices), len(directions)))]
    if return_success:
        success = [[] for _ in range(min(len(indices), len(directions)))]
    
    for i, index, res in result_list:
        if return_success:
            success[i].append(res.success)
        if return_full_results:
            result[i].append(res)
        else:
            result[i].append(res.x[index])
    
    try:
        result = np.array(result)
    except ValueError:
        result = np.array(result, dtype=object)
    
    if isScalarDirection:
        result = result.ravel()
    
    if isScalarIndex:
        result = result[0]
    
    if return_success:
        try:
            success = np.array(success)
        except ValueError:
            success = np.array(success, dtype=object)
        if isScalarDirection:
            success = success.ravel()
        if isScalarIndex:
            success = success[0]
        return result, success
    else:
        return result
    

@inherit_doc(find_CI)
def find_function_CI(x0, function, logL, 
                     functionJac = None,
                     functionHess = None,
                     logLJac = None, 
                     logLHess = None,
                     relativeError = 1e-4, 
                     **kwargs):
    """Returns the profile likelihood confidence interval(s) for a function 
    of parameters.

    Parameters
    ----------
    function : callable
       Function of the parameters for which the confidence interval shall be 
       computed
    logL : callable
        Log-likelihood function.
    functionJac : callable
        Gradient of :py:obj:`function`. If `None`, it will be computed based on
        finite differences using numdifftools.
    functionHess : callable
        Hessian of :py:obj:`function`. If `None`, it will be computed based on
        finite differences using numdifftools.
    logLJac : callable
        Gradient of :py:obj:`logL`. If `None`, it will be computed based on
        finite differences using numdifftools.
    logLHess : callable
        Hessian of :py:obj:`logL`. If `None`, it will be computed based on
        finite differences using numdifftools.
    relativeError : float
        Permitted relative error in the confidence interval bound.
    **kwargs : keyword arguments
        Other keyword arguments passed to :py:meth:`find_CI` and
        :py:meth:`find_CI_bound`. Look at the documentation there.
    """
    
    function0 = function(x0)
    error = function0 * relativeError
    x0 = np.insert(x0, 0, function0)
    
    fun_2 = lambda x: 0.5 * ((function(x[1:])-x[0])/error)**2
    fun = lambda x: logL(x[1:]) - fun_2(x)
    
    if functionJac is None and logLJac is None:
        jac = nd.Gradient(fun)
    else:
        if functionJac is None:
            jac_2 = nd.Gradient(fun_2)
        else:
            def jac_2(x):
                result = np.zeros(x0.size)
                diffTerm = (function(x[1:]) - x[0]) / error**2
                result[0] = -diffTerm
                result[1:] = np.asarray(functionJac(x[1:])) * diffTerm
                return result
        
        if logLJac is None:
            logLJac = nd.Gradient(logL)
        
        def jac(x):
            result = -jac_2(x)
            result[1:] += np.asarray(logLJac(x[1:]))
            return result
        
    if functionHess is None:
        hess_2 = nd.Hessian(fun_2)
    else:
        if functionJac is None:
            functionJac = nd.Gradient(function)
        
        def hess_2(x):
            result = np.zeros((x0.size, x0.size))
            functionJacX = np.asarray(functionJac(x[1:]) )
            result[0, 0] = 1 / error**2
            result[1:, 0] = result[0, 1:] = -functionJacX / error**2
            result[1:, 1:] = (np.asarray(functionHess(x[1:])) * (function(x[1:]) - x[0])
                              + functionJacX * functionJacX[:,None]) / error**2
            return result
        
    if logLHess is None:
        logLHess = nd.Hessian(logL)
        
    def hess(x):
        result = -hess_2(x)
        result[1:, 1:] += np.asarray(logLHess(x[1:]))
        return result
    
    # These arguments could lead to wrong results if passed on.
    # Hence, we ignore them.
    kwargs.pop("fun0", None)
    kwargs.pop("jac0", None)
    kwargs.pop("hess0", None)
    
    return find_CI(x0, fun, jac, hess, indices=0, **kwargs)


# Define aliases that adhere to the python naming conventions
find_ci = find_CI 
find_function_ci = find_function_CI 
find_ci_bound = find_CI_bound